# $Id: utils.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
"""Some various utilities."""

import sys
from os import sep, listdir, makedirs
from os.path import exists, join, isdir, dirname
import shutil
import re
import zipfile
import mimetypes
from datetime import datetime
from Crypto.Cipher import AES
from base64 import b64encode, b64decode
from cStringIO import StringIO
from beaker.crypto.util import sha1

from pyramid.i18n import TranslationStringFactory, make_localizer


_ = TranslationStringFactory('publiforge')
EXCLUDED_FILES = ('.hg', '.svn', '.git', 'Thumbs.db', '.DS_Store')
FILE_TYPES = ('css', 'csv', 'epub+zip', 'folder', 'html', 'jpeg', 'mpeg',
    'msword', 'ogg', 'pdf', 'plain', 'png', 'postscript', 'relaxng', 'tiff',
    'vnd.ms-excel', 'vnd.oasis.opendocument.spreadsheet',
    'vnd.oasis.opendocument.text', 'x-msdos-program', 'x-msvideo', 'x-python',
    'x-shockwave-flash', 'x-tar', 'x-wav', 'xml', 'xml-dtd', 'zip')
TEXT_TYPES = ('css', 'csv', 'html', 'plain', 'relaxng', 'x-python', 'xml',
    'xml-dtd')


# =============================================================================
def localizer(locale_name, directories=None):
    """Create a :class:`pyramid.i18n.Localizer` object corresponding to the
    provided locale name from the translations found in the list of translation
    directories.

    :param locale_name: (string)
        Current language.
    :param directories: (list, optional)
        Translation directories.
    :return: (:class:`pyramid.i18n.Localizer` instance)
    """
    return make_localizer(
        locale_name, directories or [join(dirname(__file__), '..', 'Locale')])


# =============================================================================
def has_permission(request, *perms):
    """Check if the user has at least one of the specified permissions.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param perms: (list)
        List of permission groups.
    :return: (boolean)

    See :ref:`frontreference_permissions`.
    """
    if not 'perms' in request.session:
        return False
    if 'admin' in request.session['perms']:
        return True
    for perm in perms:
        if perm in request.session['perms'] \
               or '%s_manager' % perm[0:3] in request.session['perms']\
               or ('%s_editor' % perm[0:3] in request.session['perms']
                   and perm[4:] == 'user'):
            return True
    return False


# =============================================================================
def config_get(config, section, option, default=None):
    """Retrieve a value from a configuration object.

    :param config: (object)
        Configuration object.
    :param section: (string)
        Section name.
    :param option: (string)
        Option name.
    :param default: (string, optional)
        Default value
    :return: (string)
        Read value or default value.
    """
    if not config.has_option(section, option):
        return default
    value = config.get(section, option)
    return (isinstance(value, str) and value.decode('utf8')) or value


# =============================================================================
def copy_content(src_dir, dst_dir, exclude=EXCLUDED_FILES):
    """Copy the content of a ``src_dir`` directory into a ``dst_dir``
    directory.

    :param src_dir: (string)
        Source directory path.
    :param dst_dir: (string)
        Destination directory path.
    :param exclude: (list)
        List of files to exclude.
    """
    if not exists(dst_dir):
        makedirs(dst_dir)
    for name in listdir(src_dir):
        if name in exclude:
            continue
        fullname = join(src_dir, name)
        if not isinstance(fullname, unicode):
            fullname = fullname.decode('utf8')
            name = name.decode('utf8')
        if isdir(fullname):
            copy_content(fullname, join(dst_dir, name), exclude)
        else:
            shutil.copy(fullname, dst_dir)


# =============================================================================
def camel_case(text):
    """Convert ``text`` in Camel Case."""
    if sys.version_info[:3] >= (2, 7, 0):
        # pylint: disable = I0011, E1123
        return re.sub(r'(^\w|[-_ 0-9]\w)',
            lambda m: m.group(0).replace('_', '').replace(' ', '').upper(),
            text, flags=re.UNICODE)
    else:
        return re.sub(r'(^\w|[-_ 0-9]\w)',
            lambda m: m.group(0).replace('_', '').replace(' ', '').upper(),
            text)


# =============================================================================
def normalize_name(name):
    """Normalize name."""
    return re.sub(r'[*?"\'/]', '_', u'_'.join(name.strip().split()))\
           .lower().encode('utf8')


# =============================================================================
def hash_sha(value, key):
    """Cryptographic hash function with SHA1 algorithm.

    :param value: (string)
        String to hash.
    :param key: (string)
        Encryption key.
    """
    return sha1('%s%s' % (value.encode('utf8'), key)).hexdigest()


# =============================================================================
def encrypt(value, key):
    """Encryption function.

    :param value: (string)
        String to encrypt.
    :param key: (string)
        Encryption key.
    :return: (string)
        Encrypted value or ``None``.
    """
    if value:
        cipher = AES.new((str(key) * 16)[:32])
        return b64encode(cipher.encrypt(value + ' ' * (16 - len(value) % 16)))


# =============================================================================
def decrypt(value, key):
    """Decryption function.

    :param value: (string)
        String to decrypt.
    :param key: (string)
        Encryption key.
    :return: (string)
        Clear value or ``None``.
    """
    if value:
        cipher = AES.new((str(key) * 16)[:32])
        return cipher.decrypt(b64decode(str(value))).strip()


# =============================================================================
def zipize(data_list, are_files):
    """Return a ZIP archive containing all data of ``data_list``.

    :param data_list: (list)
        A list of tuples such as ``(<filename_in_zip>, <string_or_filename>)``.
    :param are_files: (boolean)
        ``True`` if strings represent file name.
    :return: (ZIP)
    """
    output = StringIO()
    zip_file = zipfile.ZipFile(output, 'w', zipfile.ZIP_DEFLATED)
    for data in data_list:
        if not are_files:
            zip_file.writestr(data[0], data[1])
    zip_file.close()

    return output.getvalue()


# =============================================================================
def unzip(archive, outpath):
    """Extract an archive ignoring forbidden files.

    :param archive: (string or file)
        Name of the ZIP file.
    :param outpath: (string)
        Full path where extract the archive.
    """
    zip_file = zipfile.ZipFile(archive, 'r')
    for zipinfo in zip_file.infolist():
        if not [k for k in EXCLUDED_FILES
                if ('%s%s' % (sep, k)) in zipinfo.filename]:
            zip_file.extract(zipinfo, outpath)


# =============================================================================
def get_mime_type(filename):
    """Return the mime type of ``filename``.

    :param filename: (string)g
        File name.
    :return: (tuple)
        A tuple such as ``(mime_type, subtype)``. For instance:
        ``('text/plain', 'plain')``.
    """
    if isdir(filename):
        return 'folder', 'folder'
    mimetype = mimetypes.guess_type(filename, False)[0]
    if mimetype is None:
        return 'unknown', 'unknown'
    subtype = mimetype.partition('/')[2]
    return mimetype, subtype or mimetype


# =============================================================================
def size_label(size, is_dir):
    """Return a size in o, Kio, Mio or Gio.

    :param size: (integer)
        Size in figures.
    :param is_dir: (boolean)
        ``True`` if it is about a directory.
    :return: (string or :class:`pyramid.i18n.TranslationString`)
    """
    # For a directory
    if is_dir:
        return _('${n} items', {'n': size}) if size > 1 else \
               _('${n} item', {'n': size})

    # For a file
    if size >= 1073741824:
        return '%.1f Gio' % round(size / 1073741824.0, 1)
    elif size >= 1048576:
        return '%.1f Mio' % round(size / 1048576.0, 1)
    elif size >= 1024:
        return '%.1f Kio' % round(size / 1024.0, 1)
    return '%d o' % size


# =============================================================================
def age(mtime):
    """Return an age in minutes, hours, days or date.

    :param mtime: (datetime)
        Modification time.
    :return: (:class:`pyramid.i18n.TranslationString` or string)
        Return an age or a date if ``mtime`` is older than a year.
    """
    # pylint: disable = I0011, R0911
    if not mtime:
        return ''
    delta = datetime.now() - mtime
    if delta.days == 0 and delta.seconds < 60:
        return _('1 second') if delta.seconds <= 1 \
               else _('${s} seconds', {'s': delta.seconds})
    elif delta.days == 0 and delta.seconds < 3600:
        minutes = delta.seconds / 60
        return _('1 minute') if minutes == 1 \
               else _('${m} minutes', {'m': minutes})
    elif delta.days == 0:
        hours = delta.seconds / 3600
        return _('1 hour') if hours == 1 \
               else _('${h} hours', {'h': hours})
    elif delta.days < 7:
        return _('1 day') if delta.days == 1 \
               else _('${d} days', {'d': delta.days})
    elif delta.days < 30:
        weeks = delta.days / 7
        return _('1 week') if weeks == 1 \
               else _('${w} weeks', {'w': weeks})
    elif delta.days < 365:
        months = delta.days / 30
        return _('1 month') if months == 1 else \
               _('${m} months', {'m': months})
    return str(mtime.replace(microsecond=0))[0:-9]


# =============================================================================
def cache_key(cache, method_name, *args):
    """Compute a cache key.

    :param cache: (:class:`beaker.cache.Cache` instance)
    :param method_name: (string)
    :param args: (positional arguments)
    :return: (string)
    """
    # pylint: disable = I0011, W0141
    try:
        key = ','.join(map(str, args))
    except UnicodeEncodeError:
        key = ','.join(map(unicode, args))
    key = '%s(%s)' % (method_name, key)
    if len(key) + len(cache.namespace_name) > 250:
        key = sha1(key).hexdigest()
    return key


# =============================================================================
def cache_method():
    """Decorator to cache a method of a class with ``self.cache`` attribute.

    ``self.cache`` is a :class:`beaker.cache.Cache` instance.

    The method being decorated must only be called with positional arguments,
    and the arguments must support being stringified with ``str()``.
    """
    def _wrapper(create_method):
        """Wrapper function."""

        def _cached(self, *args):
            """Cache function."""
            if not hasattr(self, 'cache'):
                raise Exception('Class must have a "cache" attribute!')
            key = cache_key(self.cache, create_method.__name__, *args)

            def _createfunc():
                """Creation function."""
                # pylint: disable = I0011, W0142
                return create_method(self, *args)

            return self.cache.get_value(key, createfunc=_createfunc)

        return _cached

    return _wrapper



#!/usr/bin/env python
# $Id: pfpopulate.py 4526d61fab90 2012/03/11 17:47:45 patrick $
"""Populate database and storages."""

import sys
import logging
from os import listdir, getcwd, remove
from os.path import exists, join, basename, isdir
from shutil import rmtree
from optparse import OptionParser
from ConfigParser import ConfigParser
from locale import getdefaultlocale
from beaker.cache import CacheManager
from beaker.util import parse_cache_config_options
from sqlalchemy import engine_from_config

from pyramid.paster import get_appsettings, setup_logging

from ..lib.utils import _, localizer, config_get, normalize_name
from ..lib.xml import import_configuration
from ..lib.handler import HandlerManager
from ..models import ID_LEN, DBSession, Base
from ..models.users import PAGE_SIZE, User, UserIP, UserPerm
from ..models.groups import Group, GroupPerm
from ..models.storages import Storage


LOG = logging.getLogger(__name__)


# =============================================================================
def main():
    """Main function."""
    # Parse arguments
    parser = OptionParser('Usage: %prog <config_uri> [<file> <file>...]\n'
        'Examples:\n  %prog production.ini'
        '\n  %prog development.ini users.pfu.xml'
        '\n  %prog production.ini#Foo')
    parser.add_option('--drop_tables', dest='drop_tables',
        help='drop existing tables', default=False, action='store_true')
    parser.add_option('--loglevel', dest='log_level', help='log level',
        default='INFO')
    opts, args = parser.parse_args()
    log_level = getattr(logging, opts.log_level.upper(), None)
    if not log_level or len(args) < 1 or not exists(args[0].partition('#')[0]):
        parser.error('Incorrect argument')
        sys.exit(2)

    populate = Populate(args[0])
    LOG.setLevel(log_level)
    populate.all(opts.drop_tables, args[1:])


# =============================================================================
class Populate(object):
    """Class to populate database and storages."""

    # -------------------------------------------------------------------------
    def __init__(self, conf_uri):
        """Constructor method."""
        setup_logging(conf_uri)
        self._settings = get_appsettings(conf_uri)
        if len(self._settings) < 3:
            try:
                self._settings = get_appsettings(conf_uri, 'PubliForge')
            except LookupError:
                self._settings = get_appsettings(conf_uri, basename(getcwd()))
        self._config = ConfigParser({'here': self._settings['here']})
        self._config.read(self._settings['__file__'])
        self._lang = self._settings.get('pyramid.default_locale_name')

    # -------------------------------------------------------------------------
    def all(self, drop_tables, files=list()):
        """Check settings, initialize database and create storages.

        :param drop_tables: (boolean)
            If ``True`` drop existing tables before filling.
        :param files: (list, optional)
             List of files on command-line.
        """
        # Check general settings
        if not self._lang or not self._settings.get('available_languages'):
            LOG.error(self._translate(_(
                'Must define available languages and a default one.')))
            return
        if not self._settings.get('uid'):
            LOG.error(self._translate(_(
                'Must define an unique identifier for this instance.')))
            return

        # Clean build directory
        if exists(self._settings.get('build.root')):
            for name in listdir(self._settings['build.root']):
                name = join(self._settings['build.root'], name)
                if isdir(name):
                    rmtree(name)
                else:
                    remove(name)

        # Agent only
        if self._settings.get('storage.root') is None:
            return

        # Populate
        self._initialize_sql(drop_tables)
        self._populate_admin()
        self._populate_from_xml(files)
        self._update_storages()
        if not self._has_administrator():
            LOG.error(self._translate(_('You must define an administrator.')))
        DBSession.close()

    # -------------------------------------------------------------------------
    def _initialize_sql(self, drop_tables):
        """Database initialization.

        :param drop_tables: (boolean, optional)
            If ``True`` drop existing tables before filling.
        """
        # Initialize SqlAlchemy session
        engine = engine_from_config(self._settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        Base.metadata.bind = engine

        # Eventually, drop any existing tables
        if drop_tables or bool(self._settings.get('drop_tables')):
            LOG.info(self._translate(_('###### Dropping existing tables')))
            Base.metadata.drop_all()

        # Create the tables if they don't already exist
        Base.metadata.create_all(engine)

    # -------------------------------------------------------------------------
    def _populate_admin(self):
        """Populate database with an administrator."""
        # pylint: disable = I0011, W0142
        if not self._config.has_option('Populate', 'admin.login'):
            return

        LOG.info(self._translate(_('###### Adding administrator')))
        login = normalize_name(self._get('admin.login'))[0:ID_LEN]
        ips = self._get('admin.ips', '')
        record = {
            'status': 'active',
            'password': self._get('admin.password'),
            'name': self._get('admin.name'),
            'email': self._get('admin.email'),
            'lang': self._get('admin.language', self._lang),
            'restrict_ip': bool(ips),
            'home':  self._get('admin.home', 'site'),
            'page_size': int(self._get('admin.page_size', PAGE_SIZE))}
        if not record['password'] or not record['name']\
               or not record['email']:
            sys.exit('*** Incorrect administrator definition.')
        user = DBSession.query(User.login).filter_by(login=login).first()
        if user is not None:
            return
        user = User(self._settings, login, **record)
        DBSession.add(user)
        DBSession.commit()

        user.perms.append(UserPerm(user.user_id, 'all', 'admin'))
        for my_ip in ips.split():
            user.ips.append(UserIP(user.user_id, my_ip))
        DBSession.commit()

    # -------------------------------------------------------------------------
    def _populate_from_xml(self, files):
        """Populate database with XML content.

        :param files: (list)
             List of files on command-line.
        """
        if not self._config.has_section('Populate') and not files:
            return
        LOG.info(self._translate(_('###### Loading configurations')))
        for idx in range(100):
            option = 'file.%d' % idx
            if not self._config.has_option('Populate', option):
                continue
            filename = self._get(option)
            LOG.info(filename)
            errors = import_configuration(
                self._settings, filename, error_if_exists=False)
            for error in errors:
                LOG.error(self._translate(error))

        for filename in files:
            LOG.info(filename)
            errors = import_configuration(
                self._settings, filename, error_if_exists=False)
            for error in errors:
                LOG.error(self._translate(error))

    # -------------------------------------------------------------------------
    def _update_storages(self):
        """Update all storages."""
        cache_manager = CacheManager(
            **parse_cache_config_options(self._settings))
        handler_manager = HandlerManager(self._settings, cache_manager)

        LOG.info(self._translate(_('###### Updating storages')))
        for storage in DBSession.query(Storage):
            LOG.info('{0:.<32}'.format(storage.storage_id))
            storage_dir = join(
                self._settings['storage.root'], storage.storage_id)
            handler = handler_manager.get_handler(storage.storage_id, storage)
            if handler is None:
                LOG.error(self._translate(
                    _('${e}: unkwown engine', {'e': storage.vcs_engine})))
            elif not exists(storage_dir) or not listdir(storage_dir):
                error = handler.clone()
                if error:
                    LOG.error(error)
            else:
                handler.synchronize(None, True)

    # -------------------------------------------------------------------------
    @classmethod
    def _has_administrator(cls):
        """Look for an administrator.

        :return: (boolean)
        """
        # User permission
        if DBSession.query(UserPerm)\
               .filter_by(scope='all', level='admin').first():
            return True

        # Group permission
        group = DBSession.query(GroupPerm.group_id)\
                .filter_by(scope='all', level='admin').first()
        if group is not None:
            group = DBSession.query(Group).filter_by(group_id=group[0]).first()
            if len(group.users):
                return True
        return False

    # -------------------------------------------------------------------------
    def _get(self, option, default=None):
        """Retrieve a value from section [Populate].

        :param option: (string)
            Option name.
        :param default: (string, optional)
            Default value
        :return: (string)
            Read value or default value.
        """
        return config_get(self._config, 'Populate', option, default)

    # -------------------------------------------------------------------------
    @classmethod
    def _translate(cls, text):
        """Return ``text`` translated.

        :param text: (string)
            Text to translate.
        """
        return localizer(getdefaultlocale()[0]).translate(text)



# $Id: pack.py e8da66c6d933 2012/02/18 22:34:11 patrick $
# pylint: disable = I0011, C0322
"""Pack view callables."""

from colander import Mapping, SchemaNode, String, Boolean, Length

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound

from ..lib.utils import _, FILE_TYPES
from ..lib.views import get_action, current_project, file_infos
from ..lib.form import Form
from ..lib.renderer import TabSet
from ..models import LABEL_LEN, DESCRIPTION_LEN, PATH_LEN
from ..models import DBSession, close_dbsession
from ..models.projects import ProjectPack, ProjectPackFile


PACK_SETTINGS_TABS = (_('Description'), _('Files'))
FILE_TYPE_LABELS = {'file': _('Files to process'),
    'resource': _('Resource files'), 'template': _('Template files')}


# =============================================================================
class PackView(object):
    """Class to manage packs."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='pack_view',
                 renderer='../Templates/pck_view.pt')
    def view(self):
        """Display a pack."""
        project = current_project(self._request)
        pack = self._current_pack(project)
        tab_set = TabSet(self._request, PACK_SETTINGS_TABS)
        form = Form(self._request,
                    defaults={'processing_id': project['processing_id']})

        # Action
        if get_action(self._request)[0] == 'bld!' and form.validate():
            return HTTPFound(self._request.route_path(
                'build_launch', project_id=project['project_id'],
                processing_id=form.values['processing_id'],
                pack_ids=pack.pack_id))

        # Breadcrumbs
        self._request.breadcrumbs.add(
            _('Pack'), replace=self._request.route_path('pack_edit',
            project_id=project['project_id'], pack_id=pack.pack_id))

        return {
            'form': form, 'tab_set': tab_set, 'project': project, 'pack': pack,
            'files': file_infos(self._request, pack),
            'file_type_labels': FILE_TYPE_LABELS, 'FILE_TYPES': FILE_TYPES}

    # -------------------------------------------------------------------------
    @view_config(route_name='pack_create', renderer='../Templates/pck_edit.pt')
    def create(self):
        """Create a pack."""
        project = current_project(self._request)
        form, tab_set = self._settings_form()

        if form.validate():
            pack = self._create(project['project_id'], form.values)
            if pack is not None:
                self._request.breadcrumbs.pop()
                return HTTPFound(self._request.route_path('pack_edit',
                    project_id=pack.project_id, pack_id=pack.pack_id,
                    _anchor='tab1'))

        self._request.breadcrumbs.add(_('Pack creation'))
        return {
            'form': form, 'tab_set': tab_set, 'project': project, 'pack': None}

    # -------------------------------------------------------------------------
    @view_config(route_name='pack_edit', renderer='../Templates/pck_edit.pt')
    def edit(self):
        """Edit a pack."""
        # Environment
        project = current_project(self._request)
        pack = self._current_pack(project)
        form, tab_set = self._settings_form(pack)

        # Action
        # pylint: disable = I0011, E1103
        action = get_action(self._request)[0]
        if action[0:4] == 'rmv!':
            DBSession.query(ProjectPackFile).filter_by(
                project_id=project['project_id'], pack_id=pack.pack_id,
                file_type=action[4:].partition('_')[0],
                path=action[4:].partition('_')[2]).delete()
            DBSession.commit()
        elif action == 'sav!' and form.validate(pack):
            for item in pack.files:
                item.visible = form.values[
                    '%s_%s_see' % (item.file_type, item.path)]
                if item.file_type == 'template':
                    item.target = form.values['template_%s' % item.path]
            DBSession.commit()
            return HTTPFound(self._request.route_path('pack_view',
                project_id=project['project_id'], pack_id=pack.pack_id))

        # Breadcrumbs
        self._request.breadcrumbs.add(
            _('Pack'), replace=self._request.route_path('pack_view',
            project_id=project['project_id'], pack_id=pack.pack_id))
        return {
            'form': form, 'tab_set': tab_set, 'action': action,
            'project': project, 'pack': pack,
            'files': file_infos(self._request, pack),
            'file_type_labels': FILE_TYPE_LABELS, 'FILE_TYPES': FILE_TYPES}

    # -------------------------------------------------------------------------
    def _current_pack(self, project):
        """Return the current pack object.

        :param project: (dictionary)
            Current project dictionary
        :return: (:class:`~.models.projects.ProjectPack` instance)

        ``request.session['container']`` is updated with ``pack_id`` and
        ``pack_label``.
        """
        project_id = project['project_id']
        pack_id = int(self._request.matchdict.get('pack_id'))

        # Pack
        pack = DBSession.query(ProjectPack).filter_by(
            project_id=project_id, pack_id=pack_id).first()
        if pack is None:
            raise HTTPNotFound(comment=_('This pack does not exist!'))

        # Container for storage
        self._request.session['container'] = {'project_id': project_id,
            'pack_id': pack_id, 'pack_label': pack.label}

        return pack

    # -------------------------------------------------------------------------
    def _settings_form(self, pack=None):
        """Return a pack settings form.

        :param pack: (:class:`~.models.projects.ProjectPack`
            instance, optional) Current pack object.
        :return: (tuple)
             A tuple such as ``(form, tab_set)``
        """
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='label',
            validator=Length(min=3, max=LABEL_LEN)))
        schema.add(SchemaNode(String(), name='description',
            validator=Length(max=DESCRIPTION_LEN), missing=None))
        schema.add(SchemaNode(Boolean(), name='recursive', missing=False))
        defaults = {}

        if pack is not None:
            for item in pack.files:
                if item.file_type == 'template':
                    schema.add(SchemaNode(
                        String(), name='template_%s' % item.path,
                        validator=Length(max=PATH_LEN)))
                    defaults['template_%s' % item.path] = item.target
                name = '%s_%s_see' % (item.file_type, item.path)
                schema.add(SchemaNode(Boolean(), name=name, missing=False))
                defaults[name] = item.visible

        return (
            Form(self._request, schema=schema, defaults=defaults, obj=pack),
            TabSet(self._request, PACK_SETTINGS_TABS))

    # -------------------------------------------------------------------------
    def _create(self, project_id, values):
        """Create a record in ``projects_packs`` table.

        :param project_id: (string)
            Project ID.
        :param values: (dictionary)
            Values to record.
        :return:
            (:class:`~.models.projects.ProjectPack`  instance)
        """
        # Check unicity and create pack
        label = ' '.join(values['label'].split()).strip()
        if DBSession.query(ProjectPack) \
               .filter_by(project_id=project_id, label=label).first():
            self._request.session.flash(
                _('This pack already exists.'), 'alert')
            return

        # Create record
        pack = ProjectPack(project_id, label, values['description'])
        DBSession.add(pack)
        DBSession.commit()

        return pack



# $Id: test_lib_form.py 0badce09f183 2012/03/14 09:14:17 patrick $
# pylint: disable = I0011, R0904
"""Tests of ``lib.form`` classes and functions."""

import unittest
from pyramid import testing


# =============================================================================
class UnitTestLibFormButton(unittest.TestCase):
    """Unit test class for ``lib.form.button``."""

    # -------------------------------------------------------------------------
    def test_with_no_label(self):
        """[u:lib.form.button] with no label"""
        from ..lib.form import button
        from webhelpers.html import literal
        test_button = button(url='url', src='src')
        expected_result = \
            literal(u'<a href="url"><img src="src" alt="None"/></a> ')
        self.assertEqual(test_button, expected_result)

    # -------------------------------------------------------------------------
    def test_regular_button(self):
        """[u:lib.form.button] regular button"""
        from ..lib.form import button
        from webhelpers.html import literal
        test_button = \
            button(url='url', label='label',
                       src='src', title='title', class_='value')
        expected_result = \
            literal(u'<a href="url" title="title" class="value">'
                    '<img src="src" alt="label"/>label</a> ')
        self.assertEqual(test_button, expected_result)


# =============================================================================
class UnitTestLibFormValidate(unittest.TestCase):
    """Unit test class for ``lib.form.validate``."""

    # -------------------------------------------------------------------------
    def test_validate_request_no_post(self):
        """[u:lib.form.validate] request with no post"""
        from ..lib.form import Form
        request = testing.DummyRequest()
        form = Form(request)
        self.assertTrue(len(request.POST) == 0)
        self.assertFalse(form.validate())


# =============================================================================
class UnitTestLibFormBegin(unittest.TestCase):
    """Unit test class for ``lib.form.begin``."""

    # -------------------------------------------------------------------------
    def test_begin_unsecure_form(self):
        """[u:lib.form.begin] unsecure form  """
        from ..lib.form import Form
        from webhelpers.html import literal
        request = testing.DummyRequest()
        form = Form(request)
        form._secure = False
        html = form.begin(url='url')
        expected_html = literal(u'<form action="url" method="post">')
        self.assertEqual(html, expected_html)

    # -------------------------------------------------------------------------
    def test_begin_secure_form(self):
        """[u:lib.form.begin] secure form  """
        from ..lib.form import Form
        from webhelpers.html import literal
        request = testing.DummyRequest()
        form = Form(request)
        form._secure = True
        html = form.begin(url='url')
        expected_html = \
            literal('<form action="url" method="post"><div><input id="_csrf"'
                    ' name="_csrf" type="hidden" value="%s" /></div>'
                    % request.session.new_csrf_token())
        self.assertEqual(html, expected_html)



# $Id: test_scripts_pfbuild.py df43b426606b 2012/03/09 15:58:44 yvon $
# pylint: disable = I0011, R0904
"""Tests of ``scripts.pfbuild`` classes and functions."""

import unittest
from os.path import join, dirname

from ..tests import TEST_INI

ROOT = dirname(TEST_INI)
BUILD2HTML_FILE = join(ROOT, 'Examples', 'Builds', 'publidoc2html.pfb.xml')
BUILD2EPUB_FILE = join(ROOT, 'Examples', 'Builds', 'publidoc2epub.pfb.xml')
BUILD_ID = 'Test'
XML_ID = 'torture_test'


# =============================================================================
class DummyOption(object):
    """Class to simulate ``OptionParser`` option."""
    # pylint: disable = I0011, R0903

    build_id = BUILD_ID

    # -------------------------------------------------------------------------
    def __init__(self):
        """Constructor method."""
        self.list_processors = False
        self.list_variables = False
        self.show_variables = False
        self.processor_root = None
        self.storage_root = None
        self.build_root = None
        self.recursive = False


# =============================================================================
class BuildLauncherTestCase(unittest.TestCase):
    """Base unit test class for ``scripts.pfbuild.BuildLauncher``."""
    # pylint: disable = I0011, C0103

    # -------------------------------------------------------------------------
    def __init__(self, method_name='runTest'):
        """Constructor method."""
        super(BuildLauncherTestCase, self).__init__(method_name)
        self._build_file = None
        self._build_tree = None
        self._storage_root = None
        self._build_root = None
        self._opts = DummyOption()

    # -------------------------------------------------------------------------
    def setUp(self):
        """Read configuration and clean up build directory."""

        # Read configuration file
        from os.path import normpath
        from ..lib.xml import load
        self._build_tree = load(self._build_file, {'publiforge':
            join(dirname(__file__), '..', 'RelaxNG', 'publiforge.rng')})

        # Read storage path
        path = dirname(self._build_file)
        xpath = 'build/settings/setting[@key="%s"]'
        elt = self._build_tree.find(xpath % 'storage.root')
        if elt is not None:
            self._storage_root = normpath(join(path, elt.text.strip()))

        # Read build directory
        elt = self._build_tree.find(xpath % 'build.root')
        if elt is not None:
            self._build_root = normpath(join(path, elt.text.strip()))

        # Clean up build directory
        self.tearDown()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from os.path import exists
        from shutil import rmtree
        if self._build_root and exists(join(self._build_root, BUILD_ID)):
            rmtree(join(self._build_root, BUILD_ID))
        self._opts.storage_root = None
        self._opts.build_root = None

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``BuildLauncher`` object."""
        from ..scripts import pfbuild
        from ..lib.build import agent
        from logging import WARNING
        agent.LOG.setLevel(WARNING)
        pfbuild.LOG.setLevel(WARNING)
        return pfbuild.BuildLauncher(self._opts)


# =============================================================================
class UnitTestScriptsPFBuildBuildLauncherHTML(BuildLauncherTestCase):
    """Unit test class for ``scripts.pfbuild.BuildLauncher`` with HTML
    processing."""
    # pylint: disable = I0011, R0904, W0212

    # -------------------------------------------------------------------------
    def __init__(self, method_name='runTest'):
        """Constructor method."""
        super(UnitTestScriptsPFBuildBuildLauncherHTML, self).\
            __init__(method_name)
        self._build_file = BUILD2HTML_FILE

    # -------------------------------------------------------------------------
    def test_read_settings(self):
        """[u:scripts.pfbuild.BuildLauncher._read_settings:HTML]"""
        launcher = self._make_one()
        settings = launcher._read_settings(self._build_file, self._build_tree)
        self.assertTrue('storage.root' in settings)
        self.assertEqual(settings['storage.root'], self._storage_root)
        self.assertTrue('build.root' in settings)
        self.assertEqual(settings['build.root'], self._build_root)

    # -------------------------------------------------------------------------
    def test_read_settings_with_options(self):
        """[u:scripts.pfbuild.BuildLauncher._read_settings:HTML] with option"""
        from os.path import expanduser
        self._opts.storage_root = '~/Storages'
        self._opts.build_root = '~/Builds'
        launcher = self._make_one()
        settings = launcher._read_settings(self._build_file, self._build_tree)
        self.assertTrue('storage.root' in settings)
        self.assertEqual(
            settings['storage.root'], expanduser(self._opts.storage_root))
        self.assertTrue('build.root' in settings)
        self.assertEqual(
            settings['build.root'], expanduser(self._opts.build_root))

    # -------------------------------------------------------------------------
    def test_read_processing(self):
        """[u:scripts.pfbuild.BuildLauncher._read_processing:HTML]"""
        from ..lib.build.agent import AgentBuildManager
        launcher = self._make_one()
        settings = {'storage.root': self._storage_root,
                    'build.root': self._build_root}
        build_manager = AgentBuildManager(settings)
        build_manager.processor_list()
        processing = launcher._read_processing(
            settings, build_manager, self._build_tree)
        self.assertTrue('processor_id' in processing)
        self.assertTrue('variables' in processing)
        self.assertTrue('resources' in processing)
        self.assertTrue('templates' in processing)

    # -------------------------------------------------------------------------
    def test_read_pack(self):
        """[u:scripts.pfbuild.BuildLauncher._read_pack:HTML]"""
        launcher = self._make_one()
        settings = {'storage.root': self._storage_root,
                    'build.root': self._build_root}
        pack = launcher._read_pack(settings, self._build_tree, [])
        self.assertTrue('recursive' in pack)
        self.assertTrue('files' in pack)
        self.assertTrue('resources' in pack)
        self.assertTrue('templates' in pack)

    # -------------------------------------------------------------------------
    def test_read_pack_with_files(self):
        """[u:scripts.pfbuild.BuildLauncher._read_pack:HTML] with file"""
        from os.path import relpath
        launcher = self._make_one()
        settings = {'storage.root': self._storage_root,
                    'build.root': self._build_root}
        pack = launcher._read_pack(
            settings, self._build_tree, [self._build_file])
        self.assertTrue('files' in pack)
        self.assertTrue(
            relpath(self._build_file, self._storage_root) in pack['files'])

    # -------------------------------------------------------------------------
    def test_start(self):
        """[u:script.pfbuild.BuildLauncher.start:HTML]"""
        from os.path import exists
        from ..lib.utils import camel_case
        launcher = self._make_one()
        launcher.start(self._build_file, {})
        path = join(self._build_root, BUILD_ID, 'Processor')
        self.assertTrue(exists(join(path, 'processor.xml')))
        self.assertTrue(exists(join(path, 'leprisme.ini')))
        self.assertTrue(exists(join(path, 'lib')))
        self.assertTrue(exists(join(path, 'Templates')))
        self.assertTrue(exists(join(path, 'Regex')))
        self.assertTrue(exists(join(path, 'RelaxNG')))
        self.assertTrue(exists(join(path, 'Xsl', 'publidoc2html.xsl')))
        path = join(self._build_root, BUILD_ID, 'Output', camel_case(XML_ID))
        self.assertTrue(exists(join(path, 'Css')))
        self.assertTrue(exists(join(path, 'Images')))
        self.assertTrue(exists(join(path, 'RelaxNG')))
        self.assertTrue(exists(join(path, '%s.html' % XML_ID)))
        self.assertTrue(exists(join(path, '%s-tpc-1.html' % XML_ID)))


# =============================================================================
class UnitTestScriptsPFBuildBuildLauncherEPub(BuildLauncherTestCase):
    """Unit test class for ``scripts.pfbuild.BuildLauncher`` with ePub
    processing."""

    # -------------------------------------------------------------------------
    def __init__(self, method_name='runTest'):
        """Constructor method."""
        super(UnitTestScriptsPFBuildBuildLauncherEPub, self).\
            __init__(method_name)
        self._build_file = BUILD2EPUB_FILE

    # -------------------------------------------------------------------------
    def test_start(self):
        """[u:scripts.pfbuild.BuildLauncher.start:ePub]"""
        from os.path import exists
        from ..lib.utils import camel_case
        launcher = self._make_one()
        launcher.start(self._build_file, {})
        path = join(self._build_root, BUILD_ID, 'Processor')
        self.assertTrue(exists(join(path, 'processor.xml')))
        self.assertTrue(exists(join(path, 'leprisme.ini')))
        self.assertTrue(exists(join(path, 'Templates', 'Publidoc2EPub')))
        self.assertTrue(exists(join(path, 'Templates', 'Publidoc2Html')))
        self.assertTrue(exists(join(path, 'jars')))
        self.assertTrue(exists(join(path, 'Xsl', 'publidoc2html.xsl')))
        self.assertTrue(exists(join(path, 'Xsl', 'publidoc2epub.xsl')))
        path = join(self._build_root, BUILD_ID, 'Output', camel_case(XML_ID))
        self.assertTrue(exists(join(path, '%s.epub' % XML_ID)))



# $Id: build.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
# pylint: disable = I0011, C0322
"""Build view callables."""

from datetime import datetime, timedelta
from time import time
from os.path import join
from lxml import etree
from colander import SchemaNode, String, Length
import re

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound
from pyramid.response import Response
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.i18n import get_localizer

from ..lib.utils import _, normalize_name
from ..lib.xml import XML_NS, PUBLIFORGE_RNG_VERSION
from ..lib.views import get_action, current_storage, vcs_user, current_project
from ..lib.views import current_processing, variable_schema, describe_var
from ..lib.form import Form
from ..models import PATH_LEN, DBSession, close_dbsession
from ..models.projects import Project, ProjectPack


# =============================================================================
class BuildView(object):
    """Class to manage builds."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='build_launch',
                 renderer='../Templates/bld_launch.pt')
    def launch(self):
        """Launch one or more builds."""
        # Processing
        project = current_project(self._request)
        processing, processor = current_processing(self._request)
        output = processing.output and current_storage(
            self._request, processing.output.partition('/')[0])[0] or None

        # Packs
        pack_ids = self._request.matchdict.get('pack_ids').split('_')
        packs = DBSession.query(ProjectPack)\
                .filter_by(project_id=project['project_id'])\
                .filter(ProjectPack.pack_id.in_(pack_ids)).all()
        if packs is None:
            raise HTTPNotFound(comment=_('No pack to build!'))

        # Form, action & visible variables with help
        form, groups, variables = self._build_form(processing, processor)
        action, description = self._action(
            form, processing, processor, variables)

        # Launch builds
        build_ids = self._launch_builds(form, processing, processor, packs)
        if len(build_ids) == 1:
            self._request.breadcrumbs.pop()
            return HTTPFound(self._request.route_path(
                'build_progress', build_id=build_ids[0]))
        elif len(build_ids) > 1:
            self._request.breadcrumbs.pop()
            return HTTPFound(self._request.route_path(
                'project_results', project_id=project['project_id']))

        self._request.breadcrumbs.add(_('Build'))
        return {
            'form': form, 'action': action, 'project': project,
            'processing': processing, 'processor': processor, 'groups': groups,
            'variables': variables, 'description': description, 'packs': packs,
            'output': output}

    # -------------------------------------------------------------------------
    @view_config(route_name='build_view',
                 renderer='../Templates/bld_view.pt')
    def view(self):
        """Display a build, eventually, with its result."""
        # Build and project
        build, processing, processor, pack = self._current_build(True)
        project = current_project(
            self._request, int(build['build_id'].split('-')[0]))

        # Working?
        log = self._request.registry['fbuild'].progress(
            self._request, (build['build_id'],))[0]
        if log:
            return HTTPFound(self._request.route_path(
                'build_progress', build_id=build['build_id']))

        # Form & visible variables
        form, groups, variables = self._build_form(processing, processor)

        # Action
        action, description = self._action(
            form, processing, processor, variables)
        if action == 'exp!':
            return self._export_build(project, processing, processor, pack)

        # Launch build
        build_ids = self._launch_builds(form, processing, processor, (pack,))
        if build_ids:
            return HTTPFound(self._request.route_path(
                'build_progress', build_id=build_ids[0]))

        # Result
        result = self._request.registry['fbuild']\
                 .result(self._request, build['build_id'])
        log = '\n'.join([
            '%s [%-7s] %s' % (
                datetime.fromtimestamp(k[0]).isoformat(' ').split('.')[0],
                k[1], k[3]) for k in result.get('log', [])])

        self._request.breadcrumbs.add(
            _('Build'), replace=self._request.route_path(
            'build_progress', build_id=build['build_id']))
        return {
            'form': form, 'action': action, 'project': project,
            'processing': processing, 'processor': processor, 'groups': groups,
            'variables': variables, 'description': description, 'pack': pack,
            'output': build['output'], 'build': build, 'result': result,
            'message': result.get('message'), 'log': log}

    # -------------------------------------------------------------------------
    @view_config(route_name='build_progress',
                 renderer='../Templates/bld_progress.pt')
    def progress(self):
        """Display the build progress."""
        # Stop?
        build = self._current_build()[0]
        registry = self._request.registry
        is_owner = registry['fbuild'].is_owner(
            self._request, build['build_id'])
        if 'stp!' in self._request.params and is_owner:
            registry['fbuild'].stop(self._request, build['build_id'])

        # Working?
        working, prgrss = \
            registry['fbuild'].progress(self._request, (build['build_id'],))
        if not working:
            return HTTPFound(self._request.route_path(
                'build_view', build_id=build['build_id'], _anchor='build'))
        playing = str(
            timedelta(seconds=time() - prgrss[build['build_id']][3]))\
            .split('.')[0]

        # Project
        project = current_project(
            self._request, int(build['build_id'].partition('-')[0]))

        # Refresh
        if not 'ajax' in self._request.params \
               and not 'stp?' in self._request.params:
            self._request.response.headerlist.append((
                'Refresh', registry.settings.get('refresh.short', '3')))

        self._request.breadcrumbs.add(_('On progress...'))
        return {
            'form': Form(self._request), 'project': project, 'build': build,
            'is_owner': is_owner, 'progress': prgrss[build['build_id']],
            'playing': playing}

    # -------------------------------------------------------------------------
    @view_config(route_name='build_progress_ajax', renderer='json')
    def progress_ajax(self):
        """Return a JSON response for the build progress for AJAX request."""
        build_id = self._request.matchdict['build_id']
        working, prgrss = self._request.registry['fbuild']\
            .progress(self._request, (build_id,))
        playing = working and \
            str(timedelta(seconds=time() - prgrss[build_id][3])).split('.')[0]\
            or ''
        return {
            'working': working, 'percent': prgrss[build_id][1],
            'message': prgrss[build_id][2], 'playing': playing}

    # -------------------------------------------------------------------------
    @view_config(route_name='build_complete',
                 permission=NO_PERMISSION_REQUIRED)
    def complete(self):
        """Complete a build."""
        build_id = self._request.matchdict['build_id']
        key = self._request.matchdict['key']

        made = self._request.registry['fbuild']\
               .complete(self._request, build_id, key)
        return Response(made and '' or 'Unable to complete "%s"' % build_id)

    # -------------------------------------------------------------------------
    @view_config(route_name='build_log')
    def log(self):
        """Download current log."""
        build = self._current_build()[0]
        project = current_project(
            self._request, int(build['build_id'].partition('-')[0]))
        content = self._request.registry['fbuild']\
                  .result(self._request, build['build_id'])
        content = '\n'.join([
            '%s [%-7s] %s' % (
                datetime.fromtimestamp(k[0]).isoformat(' ').split('.')[0],
                k[1], k[3]) for k in content.get('log', [])])
        filename = normalize_name('%s-%s.log' % (
            project['label'], build['build_id'].partition('-')[2]))
        response = Response(content, content_type='text/plain')
        response.headerlist.append(('Content-Disposition',
            'attachment; filename="%s"' % filename))
        return response

    # -------------------------------------------------------------------------
    def _current_build(self, full=False):
        """Initialize ``request.session['build']`` and return it as current
        build dictionary.

        If ``full`` is ``True``, this method also returns
        :class:`~.models.projects.ProjectProcessing` object, processor
        :class:`lxml.etree.ElementTree` object and
        :class:`~.models.projects.ProjectPack` object.

        :param full: (boolean)
            If ``True``, force database reading and return processing and
            pack objects.
        :return: (tuple)
            A tuple such as ``(build_dictionary, processing_object,
            pack_object)`` or a :class:`pyramid.httpexceptions.`HTTPNotFound`
            exception.

        ``build_dictionary`` has the following keys: ``build_id``,
        ``processing_id``, ``processing_label``, ``varaiable_count``,
        ``pack_id``, ``pack_label``, ``file_count``, ``output``.
        """
        # Already in session
        build_id = self._request.matchdict['build_id']
        if not full and 'build' in self._request.session \
               and self._request.session['build']['build_id'] == build_id:
            return self._request.session['build'], None, None, None

        # Processing & processor
        project_id, processing_id, pack_id = build_id.split('-')
        processing, processor = \
            current_processing(self._request, project_id, processing_id)

        # Pack
        pack = DBSession.query(ProjectPack).filter_by(
            project_id=project_id, pack_id=pack_id).first()
        if pack is None:
            raise HTTPNotFound(comment=_('This pack does not exist!'))
        file_count = len([k for k in pack.files if k.file_type == 'file'])

        # Output
        output = processing.output and current_storage(
            self._request, processing.output.partition('/')[0])[0] or None
        if output:
            output = {
                'storage_id': output['storage_id'], 'label': output['label'],
                'path': processing.output.partition('/')[2]}

        self._request.session['build'] = {
            'build_id': build_id, 'processing_id': processing_id,
            'processing_label': processing.label,
            'variable_count': len(processing.variables), 'pack_id': pack_id,
            'pack_label': pack.label, 'file_count': file_count,
            'output': output}
        return self._request.session['build'], processing, processor, pack

    # -------------------------------------------------------------------------
    def _build_form(self, processing, processor):
        """Return a build form with a validating schema for visible variables.

        :param processing:
            (:class:`~.models.projects.ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :return: (tuple)
            A tuple such as ``(form, groups, variables)``

        ``groups`` is a list of group :class:`lxml.etree.Element` objects
        containing visible variables. ``variables`` is a dictionary of
        :class:`~.models.projects.ProjectProcessingVariable` objects of visible
        variables.
        """
        variables = dict([
            (k.name, k) for k in processing.variables if k.visible])
        groups = []
        for group in processor.findall('processor/variables/group'):
            for var in group.findall('var'):
                if var.get('name') in variables:
                    groups.append(group)
                    break

        schema, defaults = variable_schema(processor, variables, True, False)
        schema.add(SchemaNode(String(), name='path',
            validator=Length(max=PATH_LEN), missing=None))
        defaults['path'] = processing.output.partition('/')[2]
        form = Form(
            self._request, schema=schema, defaults=defaults, obj=processing)
        return form, groups, variables

    # -------------------------------------------------------------------------
    def _launch_builds(self, form, processing, processor, packs):
        """Launch build.

        :param form: (:class:`~.lib.form.Form` instance)
             Build form.
        :param processing:
             (:class:`~.models.projects.ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param packs: (list)
             List of :class:`~.models.projects.ProjectPack` objects.
        :return: (list)
             List of launched build
        """
        if not 'bld!.x' in self._request.params \
               or not form.validate(obj=processing):
            return tuple()
        processing.output = processing.output.partition('/')[0] \
            + '/' + (form.values['path'] or '')

        # Save variable modifications
        for var in processing.variables:
            if var.visible:
                var.value = form.values[var.name]
        DBSession.commit()

        # Create builds and start them
        build_ids = []
        for pack in packs:
            front_build = self._request.registry['fbuild'].start_build(
                self._request, processing, processor, pack)
            if front_build is not None \
                   and not self._request.session.peek_flash('alert'):
                build_ids.append(front_build.uid)

        return build_ids

    # -------------------------------------------------------------------------
    def _action(self, form, processing, processor, variables):
        """Return current action and a description text.

        :param form: (:class:`~.lib.form.Form` instance)
            Current form.
        :param processing:
             (:class:`~.models.projects.ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param variables: (dictionary)
            Variables of current processing.
        :return: (tuple)
            A tuple such as ``(action, description)``.
        """
        description = None
        action = get_action(self._request)[0]

        if action[0:4] == 'des!':
            description = describe_var(
                self._request, processor, variables, action[4:])
        elif action[0:4] == 'rst!':
            name = action[4:]
            if name in variables:
                form.values[name] = variables[name].default
                form.static(name)
        elif action[0:4] == 'upl!':
            action = action[4:].split('/')
            self._file_upload(action[0], action[1:], processing.label)
            action = ''

        return action, description

    # -------------------------------------------------------------------------
    def _export_build(self, project, processing, processor, pack):
        """Export build.

        :param project: (dictionary)
            Project dictionary.
        :param processing:
            (:class:`~.models.projects.ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param pack: (:class:`~.models.projects.ProjectPack` instance)
        :return: (:class:`pyramid.response.Response` instance)
        """
        # Header
        settings = self._request.registry.settings
        root = etree.Element('publiforge', version=PUBLIFORGE_RNG_VERSION)
        build_elt = etree.SubElement(root, 'build')
        build_elt.set('%sid' % XML_NS, '%s_%d-%d-%d' % (settings['uid'],
            project['project_id'], processing.processing_id, pack.pack_id))

        # Settings
        build_elt.append(etree.Comment('=' * 68))
        group_elt = etree.SubElement(build_elt, 'settings')
        etree.SubElement(group_elt, 'setting', key='storage.root')\
            .text = '../../Storages~'
        etree.SubElement(group_elt, 'setting', key='build.root')\
            .text = '../../Builds~'
        etree.SubElement(group_elt, 'setting', key='build.develop')\
            .text = settings.get('build.develop', 'true')
        etree.SubElement(group_elt, 'setting', key='processor.roots')\
            .text = '../../Processors'

        # Processing
        build_elt.append(etree.Comment('=' * 68))
        group_elt = etree.SubElement(build_elt, 'processing')
        etree.SubElement(group_elt, 'processor').text = processing.processor
        Project.xml_variables(group_elt, processing, processor, False)
        Project.xml_file_set(group_elt, processing, 'resource')
        Project.xml_file_set(group_elt, processing, 'template')

        # Pack
        build_elt.append(etree.Comment('=' * 68))
        group_elt = etree.SubElement(build_elt, 'pack')
        if pack.recursive:
            group_elt.set('recursive', 'true')
        Project.xml_file_set(group_elt, pack, 'file')
        Project.xml_file_set(group_elt, pack, 'resource')
        Project.xml_file_set(group_elt, pack, 'template')

        # Response
        filename = normalize_name('%s-%d-%d.pfb.xml' % (
            project['label'], processing.processing_id, pack.pack_id))
        xml = etree.tostring(
            root, pretty_print=True, encoding='utf-8', xml_declaration=True)
        xml = re.sub('\s*<(/?)value>\s*', '<\\1value>', xml)
        response = Response(xml, content_type='application/xml')
        response.headerlist.append(('Content-Disposition',
                'attachment; filename="%s"' % filename))
        return response

    # -------------------------------------------------------------------------
    def _file_upload(self, storage_id, path, processing_label):
        """Update a file in a storage.

        :param storage_id: (string)
            Storage ID.
        :param path: (list)
            Splitted relative path of the file to update.
        :param processing_label: (string)
            Processing label.
        """
        upload_file = self._request.params.get('upload_file')
        if upload_file is None or isinstance(upload_file, basestring)\
               or not path:
            return

        storage, handler = current_storage(self._request, storage_id)
        if storage['perm'] == 'user':
            self._request.session.flash(_(
                "You can't modify this storage!"), 'alert')
            return
        if path[-1] != upload_file.filename:
            self._request.session.flash(
                _('File names are different.'), 'alert')
            return

        message = get_localizer(self._request).translate(
            _('Updated for "${p}"', {'p': processing_label}))
        handler.upload(vcs_user(
            self._request, storage), join('.', *path[0:-1]),
            upload_file, path[-1], message)
        if handler.progress()[0] == 'error':
            self._request.session.flash(handler.progress()[1], 'alert')
        else:
            self._request.session.flash(_('File has been updated.'))



# $Id: none.py 513105aef084 2012/02/13 16:15:59 patrick $
"""Storage with no Version Control System management."""

from os import makedirs, remove, renames
from os.path import exists, isfile, getmtime
from shutil import rmtree
from datetime import datetime

from pyramid.i18n import TranslationString

from ...lib.utils import _
from ...lib.vcs import Vcs


# =============================================================================
class VcsNone(Vcs):
    """No Version Control System."""

    engine = 'none'

    # -------------------------------------------------------------------------
    def clone(self, handler=None):
        """Create a directory.

        See abstract function :meth:`~.lib.vcs.Vcs.clone`.
        """
        if not exists(self.path):
            makedirs(self.path)

    # -------------------------------------------------------------------------
    def pull_update(self, handler):
        """Do nothing.

        See abstract function :meth:`~.lib.vcs.Vcs.pull_update`.
        """
        pass

    # -------------------------------------------------------------------------
    def commit_push(self, message, user_id, password, name, handler):
        """Do nothing.

        See abstract function :meth:`~.lib.vcs.Vcs.commit_push`.
        """
        pass

    # -------------------------------------------------------------------------
    def revert_all(self, handler):
        """Revert all files of the repository."""
        pass

    # -------------------------------------------------------------------------
    def last_change(self):
        """Return the last change on the repository.

        See :meth:`~.lib.vcs.Vcs.last_change`.
        """
        return datetime.fromtimestamp(getmtime(self.path)), '-', '-'

    # -------------------------------------------------------------------------
    def file_infos(self, path, filename):
        """Return information on a file.

        See :meth:`~.lib.vcs.Vcs.last_change`.
        """
        return (datetime.fromtimestamp(
            getmtime(self.full_path(path, filename))), '-', '-', '-')

    # -------------------------------------------------------------------------
    def log(self, path, filename, limit=1):
        """show revision history of file ``filename``.

        See :meth:`~.lib.vcs.Vcs.log`.
        """
        return ((datetime.fromtimestamp(
            getmtime(self.full_path(path, filename))), '-', '-', '-'),)

    # -------------------------------------------------------------------------
    def add(self, path, handler):
        """Do nothing.

        See abstract function :meth:`~.lib.vcs.Vcs.add`.
        """
        pass

    # -------------------------------------------------------------------------
    def rename(self, path, filename, new_name, handler):
        """Rename a file."""
        new_name = self.full_path(path, new_name)
        if isinstance(new_name, TranslationString):
            return handler.report('error', new_name)
        filename = self.full_path(path, filename)
        if isinstance(filename, TranslationString):
            return handler.report('error', filename)
        if exists(new_name):
            return handler.report('error', _('File already exists!'))
        try:
            renames(filename, new_name)
        except OSError, error:
            return handler.report('error', error)

    # -------------------------------------------------------------------------
    def remove(self, path, filename, handler):
        """Remove a file.

        See :meth:`~.lib.vcs.Vcs.remove`.
        """
        filename = self.full_path(path, filename)
        if isinstance(filename, TranslationString):
            return handler.report('error', filename)
        if exists(filename):
            if isfile(filename):
                remove(filename)
            else:
                rmtree(filename)

    # -------------------------------------------------------------------------
    def revision(self, fullname, revision):
        """Retrieve a revision.

        See :meth:`~.lib.vcs.Vcs.revision`.
        """
        with open(fullname, 'rb') as hdl:
            content = hdl.read()
        return content

    # -------------------------------------------------------------------------
    def diff(self, fullname, revision):
        """Return differences between revision ``revision`` and current
        revision.

        See :meth:`~.lib.vcs.Vcs.diff`.
        """
        return ''



# $Id: xml.py 4526d61fab90 2012/03/11 17:47:45 patrick $
"""XML manipulation."""

from os.path import join, dirname,  basename
from lxml import etree
from cStringIO import StringIO

from pyramid.httpexceptions import HTTPForbidden
from pyramid.response import Response

from .utils import _, has_permission, normalize_name
from ..models.users import User
from ..models.groups import Group
from ..models.storages import Storage
from ..models.projects import Project


XML_NS = '{http://www.w3.org/XML/1998/namespace}'
PUBLIFORGE_RNG_VERSION = '1.0'


# =============================================================================
def load(filename, relaxngs=None, data=None):
    """Load an XML document and validate it against a Relax NG file.

    :param filename: (string)
        Name of XML file.
    :param relaxngs: (dictionary, optional)
        Relax NG dictionary such as ``{<pattern>: <relax_ng_file>,...}``. If it
        is ``None``, no validation is performed.
    :param data: (string or :class:`ElementTree`, optional)
        Content of the XML document. If it is not ``None``, it is used in place
        of the content of the file ``filename``.
    :return: (string, :class:`TranslationString` or :class:`ElementTree`)
        An error message or an instance of :class:`lxml.etree.ElementTree`
        class.
    """
    # pylint: disable = I0011, R0911
    # Read file
    if data is None or isinstance(data, basestring):
        try:
            tree = etree.parse(filename if data is None else StringIO(data))
        except IOError, err:
            return str(err).replace('%s/' % dirname(filename), '')
        except etree.XMLSyntaxError, err:
            return _('${name} not well-formed: ${err}',
                     {'name': basename(filename), 'err': str(err)})
    else:
        tree = data

    # Find RelaxNG
    if relaxngs is None:
        return tree
    if tree.getroot().tag not in relaxngs:
        return _('${tag}: Relax NG not found', {'tag': tree.getroot().tag})
    relaxng = relaxngs[tree.getroot().tag]

    # Load RelaxNG
    if isinstance(relaxng, basestring):
        try:
            relaxng = etree.RelaxNG(etree.parse(relaxng))
        except IOError, err:
            return str(err).replace('%s/' % dirname(filename), '')
        except (etree.XMLSyntaxError, etree.RelaxNGParseError), err:
            return '"%s": %s' % (relaxng, err)

    # Validate
    if not relaxng.validate(tree):
        return str(relaxng.error_log.last_error)\
               .replace('%s/' % dirname(filename), '')\
               .replace(':ERROR:RELAXNGV', '')

    return tree


# =============================================================================
def local_text(root_elt, xpath, request=None, lang=None, default=''):
    """Return the text in local language of the ``root_elt`` element child
    selected by ``xpath``.

    :param root_elt: (:class:`lxml.etree.Element` instance)
        Root element.
    :param xpath: (string)
        XPath expression.
    :param request: (:class:`pyramid.request.Request` instance, optional)
        Current request.
    :param lang: (string, optional)
        Preferred language.
    :param default: (string, optional)
        Default label.
    :return: (string)

    If label does not exist in the asked language, this method returns the
    label in default language or in english or the first label or ''.
    """
    if lang is None and request is None:
        return default

    lang = lang or request.session['lang']
    lang_xpath = xpath + '[@xml:lang="%s"]'
    text = root_elt.xpath(lang_xpath % lang) \
           or root_elt.xpath(lang_xpath % lang.split('_')[0])
    if not len(text) and request is not None:
        lang = request.registry.settings['pyramid.default_locale_name']
        text = root_elt.xpath(lang_xpath % lang) \
               or root_elt.xpath(lang_xpath % lang.split('_')[0])
    if not len(text):
        text = root_elt.xpath(lang_xpath % 'en')

    text = (len(text) == 1 and text[0].text) or \
           (root_elt.find(xpath) and root_elt.find(xpath).text) or default

    return ' '.join(text.split()).strip()


# =============================================================================
def import_configuration(settings, filename, only=None, error_if_exists=True,
                         xml=None, request=None):
    """Import a XML configuration file.

    :param settings: (dictionary)
        Application settings
    :param filename: (string)
        Full path to file to import.
    :param only: (string, optional)
        ``user``, ``group``, ``storage`` or ``project``.
    :param error_if_exists: (boolean, default=True)
        It returns an error if an item already exists.
    :param xml: (string, optional)
        Content of the XML document.
    :param request: (:class:`pyramid.request.Request` instance, optional)
        Current request.
    :return: (list)
        A list of error messages.
    """
    # Load XML
    tree = load(filename, {'publiforge':
        join(dirname(__file__), '..', 'RelaxNG', 'publiforge.rng')}, xml)
    if isinstance(tree, basestring):
        return (tree,)

    # Load users
    # pylint: disable = I0011, E1103
    errors = []
    if only is None or only == 'user':
        for elt in tree.xpath('user|users/user'):
            errors.append(User.load(settings, elt, error_if_exists))

    # Load groups
    if only is None or only == 'group':
        for elt in tree.xpath('group|groups/group'):
            errors.append(Group.load(elt, error_if_exists))

    # Load storages
    if only is None or only == 'storage':
        for elt in tree.xpath('storage|storages/storage'):
            storage = Storage.load(settings, elt, error_if_exists)
            if isinstance(storage, basestring):
                errors.append(storage)
            elif storage is not None and request is not None:
                handler = request.registry['handler']\
                      .get_handler(storage.storage_id, storage)
                handler.clone(request)

    # Load projects
    if only is None or only == 'project':
        for elt in tree.xpath('project|projects/project'):
            errors.append(Project.load(elt, error_if_exists))

    return [k for k in errors if k is not None]


# =============================================================================
def upload_configuration(request, perm, only=None):
    """Upload a XML configuration file.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param perm: (string)
        Level of permission needed.
    :param only: (string, optional)
        ``user``, ``group``, ``storage`` or ``project``.
    """
    if not has_permission(request, perm):
        raise HTTPForbidden()

    xml_file = request.params.get('xml_file')
    if isinstance(xml_file, basestring):
        return

    errors = import_configuration(
        request.registry.settings, xml_file.filename, only=only,
        xml=xml_file.file.read(), request=request)
    for error in errors:
        request.session.flash(error, 'alert')


# =============================================================================
def export_configuration(elements, filename=None, command=False):
    """Export an XML configuration and return it as a
    :class:`pyramid.response.Response` object.

    :param elements: (list)
        List of :class:`lxml.etree.Element` objects.
    :param filename: (string, optional)
        Name of file to export. Default to ``'publiforge'``.
    :param command: (boolean, default=False)
        ``True`` if called by command line.
    :return: (:class:`pyramid.response.Response` instance)
    """
    def _label(elt):
        """Get label or name of ``elt``."""
        label = elt.get('%sid' % XML_NS) or elt.get('login') \
            or (elt.findtext('label') is not None and elt.findtext('label'))
        return u' '.join(label.strip().split())

    # Nothing to do
    if len(elements) == 0:
        raise HTTPForbidden()

    # Create XML document
    root = etree.Element('publiforge', version=PUBLIFORGE_RNG_VERSION)

    # Single export
    if len(elements) == 1:
        label = _label(elements[0])
        root.append(etree.Comment('=' * 70))
        root.append(etree.Comment(u'{0:^70}'.format(label)))
        root.append(etree.Comment('=' * 70))
        root.append(elements[0])
        filename = '%s.pf%s' % (label, elements[0].tag[0])
    # Multiple export
    else:
        for name in ('user', 'group', 'storage', 'project', 'processing'):
            elts = [k for k in elements if k is not None and k.tag == name]
            if not len(elts):
                continue
            root.append(etree.Comment('=' * 70))
            root.append(etree.Comment(u'{0:^70}'.format(
                '%ss' % name.capitalize())))
            root.append(etree.Comment('=' * 70))
            grouper = etree.SubElement(root, '%ss' % name)
            for elt in elts:
                grouper.append(etree.Comment(' ' * 68))
                grouper.append(etree.Comment(
                    u'{0:=^68}'.format(' %s ' % _label(elt))))
                grouper.append(etree.Comment(' ' * 68))
                grouper.append(elt)
        filename = filename or 'publiforge'

    # Command line
    if command:
        root = etree.ElementTree(root)
        root.write(filename, pretty_print=True, encoding='utf-8',
                    xml_declaration=True)
        return

    # Response
    response = Response(body=etree.tostring(
        root, pretty_print=True, encoding='utf-8', xml_declaration=True),
        content_type='application/xml')
    response.headerlist.append(('Content-Disposition',
        'attachment; filename="%s.xml"' % normalize_name(filename)))
    return response



# $Id: iniscript.py b1ffa5a75b89 2012/02/19 23:23:01 patrick $
"""INI scripts management."""

from os import walk
from os.path import join, exists, dirname, basename, splitext, normpath
from os.path import getmtime
from ConfigParser import ConfigParser
from subprocess import Popen, PIPE

from ...utils import config_get


MEDIA_EXT = {'image': 'svg png tif tiff jpg jpeg eps gif',
             'audio': 'flac wav ogg ac3 mp3',
             'video': 'dv mpg mov avi flv'}


# =============================================================================
class IniScript(object):
    """Class for INI script managment."""

    # -------------------------------------------------------------------------
    def __init__(self, engine):
        """Constructor method.

        :param engine: (:class:`~.lib.engine.leprisme.Engine` instance)
            Engine object on which it depends.
        """
        self._engine = engine

    # -------------------------------------------------------------------------
    def execute(self, output, filename, ini_file):
        """Make an ePub file.

        :param output: (string)
            Path to the output directory.
        :param filename: (string)
            Relative path to the original file to transform.
        :param ini_file: (string)
            Path to INI file.
        """
        # Read INI file
        config = ConfigParser({
            'here': dirname(ini_file), 'output': output,
            'stgpath': self._engine.build.data_path,
            'filepath': join(self._engine.build.data_path, dirname(filename)),
            'id': '{id}', 'ext': '{ext}', 'source': '{source}',
            'target': '{target}'})
        config.read(ini_file)

        # Compute full path to source
        data_id = config_get(config, 'Source', 'id')
        source = config_get(config, 'Source', 'file')
        if source:
            source = source.format(id=data_id)
        if not source or not exists(source):
            data_type = config_get(config, 'Source', 'type')
            if data_type in ('image', 'audio', 'video'):
                source = self._find_media(data_type, data_id,
                    config_get(config, 'Source', 'paths'),
                    config_get(config, 'Source', 'patterns'))
        if not source or not exists(source):
            self._engine.build.stopped('%s: no source for "%s"'
                % (basename(ini_file), data_id), 'a_error')
            return

        # Read full path to target
        target = config_get(config, 'Target', 'file')
        if not target:
            self._engine.build.stopped('%s: no target' % basename(ini_file))
            return

        # Already processed?
        if not self._engine.build.processing['variables'].get('force') \
               and exists(target) and getmtime(target) > getmtime(source):
            return

        # Transform data
        section = 'Transformation:%s' % splitext(source)[1][1:]
        if not config.has_section(section):
            section = 'Transformation'
        for step in range(100):
            cmd = config_get(config, section, 'step.%d' % step)
            if cmd:
                cmd = cmd.format(source=source, target=target).split()
                try:
                    err = Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()[1]
                except OSError, err:
                    self._engine.build.stopped(err)
                if not exists(target):
                    self._engine.build.stopped(err, 'a_error')
                    return
            if self._engine.build.stopped():
                return

    # -------------------------------------------------------------------------
    def _find_media(self, media_type, media_id, sources, patterns):
        """Look for a media.

        :param media_type: (string)
            Type of media (image, audio or video).
        :param media_id: (string)
            Media ID.
        :param sources: (string)
            List of source paths to browse.
        :param patterns: (string)
            List of patterns to check.
        :return: (string)
            Full path to source file or ``None``.
        """
        # pylint: disable = I0011, R0912
        if not media_id or not sources or not patterns:
            return

        # Get extension, path and pattern list
        extensions = self._engine.config(
            'Input', '%s_ext' % media_type, MEDIA_EXT[media_type]).split()
        sources = [normpath(k.strip()) for k in sources.split(',')]
        patterns = [k.strip() for k in patterns.split(',')]
        done = []

        # Local function
        def look_in_directory(source):
            """Look for a source in a directory."""
            for pattern in patterns:
                if '{ext}' not in pattern:
                    name = join(source, pattern.format(id=media_id))
                    if exists(name):
                        return name
                    continue
                for ext in extensions:
                    name = join(source, pattern.format(id=media_id, ext=ext))
                    if exists(name):
                        return name

        # Look for media in sources
        for source in sources:
            name = look_in_directory(source)
            if name is not None:
                return name
            for path, dirs, name in walk(source):
                for name in dirs:
                    name = normpath(join(path, name))
                    if name in done:
                        continue
                    done.append(name)
                    name = look_in_directory(name)
                    if name is not None:
                        return name
            done.append(source)



# $Id: test_models_storages.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``models.storages`` classes."""

import unittest
from tempfile import mkdtemp

from ..tests import ModelTestCase


# =============================================================================
class UnitTestModelsStoragesStorage(ModelTestCase):
    """Unit test class for ``models.storages.Storage``."""
    # pylint: disable = I0011, C0103, R0904

    _key = 'seekrit'
    _storage_id = '_test_storage'
    _storage_root = mkdtemp()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from os.path import exists
        from shutil import rmtree
        from ..models.storages import Storage
        if self.session.query(Storage)\
               .filter_by(storage_id=self._storage_id).first() is not None:
            Storage.delete(self._storage_root, self._storage_id)
        self.session.commit()
        if exists(self._storage_root):
            rmtree(self._storage_root)
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make an ``Storage`` object."""
        from ..models.storages import Storage
        return Storage(
            {'auth.key': self._key}, self._storage_id, 'Test storage')

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.storages.Storage.__init__]"""
        from ..models.storages import Storage
        storage1 = self._make_one()
        self.session.add(storage1)
        self.session.commit()
        storage2 = self.session.query(Storage)\
                     .filter_by(storage_id=storage1.storage_id).first()
        self.assertNotEqual(storage2, None)
        self.assertEqual(storage2.storage_id, storage1.storage_id)
        self.assertEqual(storage2.access, storage1.access)
        self.assertEqual(storage2.vcs_engine, storage1.vcs_engine)
        self.assertEqual(storage2.refresh, storage1.refresh)

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.storages.Storage.__repr__]"""
        storage = self._make_one()
        info = repr(storage)
        self.assertTrue('<Storage(' in info)
        self.assertTrue("'%s'" % storage.storage_id in info)
        self.assertTrue("'%s'" % storage.access in info)
        self.assertTrue("'%s'" % storage.vcs_engine in info)
        self.assertTrue(str(storage.refresh) in info)

    # -------------------------------------------------------------------------
    def test_delete(self):
        """[u:models.storages.Storage.delete]"""
        from os.path import exists, join
        from ..models.storages import Storage
        storage = self._make_one()
        storage_id = storage.storage_id
        self.session.add(storage)
        self.session.commit()
        storage.delete(self._storage_root, storage.storage_id)
        self.session.commit()
        storage = self.session.query(Storage)\
                    .filter_by(storage_id=storage_id).first()
        self.assertEqual(storage, None)
        self.assertFalse(exists(join(self._storage_root, storage_id)))


# =============================================================================
class UnitTestModelsStoragesStorageUser(unittest.TestCase):
    """Unit test class for ``models.storages.StorageUser``."""

    _key = 'seekrit'
    _storage_id = '_test_storage'
    _user_id = 1000

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``StorageUser`` object."""
        from ..models.storages import StorageUser
        return StorageUser(
            self._storage_id, self._user_id, True, 'user', 'user1',
            'sesame', {'auth.key': self._key})

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.storages.StorageUser.__init__]"""
        from ..lib.utils import encrypt
        storage_user = self._make_one()
        self.assertEqual(storage_user.user_id, self._user_id)
        self.assertTrue(storage_user.in_menu)
        self.assertEqual(storage_user.perm, 'user')
        self.assertEqual(storage_user.vcs_user, 'user1')
        self.assertEqual(
            storage_user.vcs_password, encrypt('sesame', self._key))

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.storages.StorageUser.__repr__]"""
        storage_user = self._make_one()
        info = repr(storage_user)
        self.assertTrue('<StorageUser(' in info)
        self.assertTrue('%s' % storage_user.storage_id in info)
        self.assertTrue('%s' % storage_user.user_id in info)


# =============================================================================
class IntegrationTestModelsStorages(ModelTestCase):
    """Integration test class for ``models.storages``."""
    # pylint: disable = I0011, C0103

    _key = 'seekrit'
    _user_id = 1000
    _storage_id = '_test_storage'
    _storage_root = mkdtemp()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from os.path import exists
        from shutil import rmtree
        from ..models.users import User
        from ..models.storages import Storage
        self.session.query(User).filter_by(user_id=self._user_id).delete()
        if self.session.query(Storage)\
               .filter_by(storage_id=self._storage_id).first() is not None:
            Storage.delete(self._storage_root, self._storage_id)
        self.session.commit()
        if exists(self._storage_root):
            rmtree(self._storage_root)
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_storage(self):
        """Make a ``Storage`` object."""
        # pylint: disable = I0011, E1101
        from ..models.storages import Storage
        return Storage(
            {'auth.key': self._key}, self._storage_id, 'Test storage')

    # -------------------------------------------------------------------------
    def _make_user(self):
        """Make an ``User`` object."""
        from ..models.users import User
        user = User({'auth.key': self._key}, '_test_user', status='active',
                    password='mypassword', name=u'Gédéon TEUSEMANIE',
                    email='test@prismallia.fr', lang='en')
        user.user_id = self._user_id
        return user

    # -------------------------------------------------------------------------
    def test_delete_user(self):
        """[i:models.storages] remove user"""
        from ..models.users import User
        from ..models.storages import StorageUser
        storage = self._make_storage()
        user = self._make_user()
        self.session.add(user)
        storage.users.append(
            StorageUser(self._storage_id, self._user_id))
        self.session.add(storage)
        self.session.commit()
        self.assertEqual(len(storage.users), 1)
        self.session.query(User).filter_by(user_id=self._user_id).delete()
        self.session.commit()
        self.assertEqual(len(storage.users), 0)



# $Id: agent.py 87b7d52fda37 2012/02/23 17:08:44 patrick $
"""Agent build management.

An *agent* is a Web service.
"""

import logging
from os import listdir, makedirs
from os.path import exists, join, getatime, normpath, dirname
from lxml import etree
from threading import Thread
from time import time
from fnmatch import fnmatch
import shutil
import urllib

from pyramid.asset import abspath_from_asset_spec

from ..utils import _, localizer, copy_content, camel_case
from ..xml import load


LOG = logging.getLogger(__name__)


# =============================================================================
class AgentBuildManager(object):
    """This class manages agent builds.

    One instance of :class:`AgentBuildManager` is created during application
    initialization. It is only used in agent mode. It is stored in application
    registry.

    ``self._processors`` is a tuple such as ``(processor_dictionary, path_list,
    available_list)``. ``processor_dictionary`` is a dictionary such as
    ``{processor_id: processor_path,...}``.

    ``self._fronts`` is a dictionary such as ``{front_id: password,...}``.

    ``self._builds`` is a dictionary of :class:`AgentBuild` objects.

    ``self._results`` is a dictionary of dictionaries such as ``{build_id:
    result_dict}``. ``result_dict`` is a dictionary with following keys:
    ``status``, ``log``, ``expire``. According to build events, it can also
    contains ``files``, ``values`` and ``error`` keys.

    ``self._results[build_id]['status']`` is one of the following strings:
    ``a_stop``, ``a_fatal`` or ``a_end``.

    ``self._results[build_id]['log']`` is a list of tuples such as
    ``(timestamp, step, percent, message)``.
    """
    # pylint: disable = I0011, R0902

    # -------------------------------------------------------------------------
    def __init__(self, settings):
        """Constructor method.

        :param settings: (dictionary)
            Setting dictionary.
        """
        # Attributes
        self.develop = bool(settings.get('build.develop') == 'true')
        self.buildspace_root = settings.get('buildspace.root')
        self.storage_root = settings.get('storage.root')
        self.build_root = settings['build.root']
        self.build_ttl = int(settings.get('build.ttl', 1800))
        self.result_ttl = int(settings.get('build.result_ttl', 86400))
        self._concurrent = int(settings.get('build.concurrent', 3))
        self._buildspace_ttl = int(settings.get('buildspace.ttl', 2678400))
        self._builds = {}
        self._results = {}

        # Processor list
        self._processors = [
            {}, tuple(settings.get('processor.roots', '').split()),
            tuple(settings.get('processor.available', '').split())]

        # Authorized front list
        self._fronts = {}
        for idx in range(100):
            if 'front.%d.uid' % idx in settings:
                self._fronts[settings['front.%d.uid' % idx]] = \
                       settings.get('front.%d.password' % idx, '')

    # -------------------------------------------------------------------------
    def processor_list(self):
        """Refresh information and return a list of available processors."""
        self._processors[0] = {}
        self.add_processors(join(dirname(__file__), '..', '..', 'Processors'))
        for path in self._processors[1]:
            self.add_processors(path)
        plist = []
        for pid in self._processors[0]:
            for pattern in self._processors[2]:
                if pid not in plist and fnmatch(pid, pattern):
                    plist.append(pid)
                    break
        return plist

    # -------------------------------------------------------------------------
    def add_processors(self, path):
        """Add all processors in path ``path``.

        :param path: (string)
             Where to look for processors.
        """
        path = abspath_from_asset_spec(path)
        if not exists(path):
            return

        for pid in listdir(path):
            if exists(join(path, pid, 'processor.xml')):
                self._processors[0][pid] = normpath(join(path, pid))

    # -------------------------------------------------------------------------
    def processor_path(self, processor_id):
        """Return processor path if exists.

        :param processor_id: (string)
            Processor ID.
        :return: (string)
        """
        return self._processors[0].get(processor_id)

    # -------------------------------------------------------------------------
    def processor_xml(self, processor_id):
        """Return processor XML if exists.

        :param processor_id: (string)
            Processor ID.
        :return: (string)
        """
        if not processor_id in self._processors[0]:
            return ''
        tree = load(
            join(self._processors[0][processor_id], 'processor.xml'),
            {'publiforge': join(dirname(__file__), '..', '..', 'RelaxNG',
             'publiforge.rng')})
        if isinstance(tree, basestring):
            return ''
        etree.strip_elements(tree, etree.Comment)
        return ' '.join(etree.tostring(tree, encoding='utf8').split())

    # -------------------------------------------------------------------------
    def front_list(self):
        """Return a list of authorized fronts."""
        return self._fronts.keys()

    # -------------------------------------------------------------------------
    def authorized_front(self, front_id, password):
        """``True`` if ``front_id`` is authorized to use agent services."""
        return front_id in self._fronts and self._fronts[front_id] == password

    # -------------------------------------------------------------------------
    def activity(self):
        """Return the global activity i.e. the number of active builds."""
        return len(self._builds)

    # -------------------------------------------------------------------------
    def start_build(self, build_id, context, processing, pack, end_url=None):
        """Create a build, add it in ``self._builds`` dictionary and try to
        start it.

        :param build_id: (string)
            Build ID.
        :param context: (dictionary)
            See :class:`~.front.FrontBuildManager`
            :meth:`~.front.FrontBuildManager.call` method.
        :param processing: (dictionary)
            A processing dictionary.
        :param pack: (dictionary)
            A pack dictionary.
        :param end_url: (string, optional)
            URL to call to complete the build.
        :return: (:class:`~AgentBuild`)
        """
        self._cleanup()
        self._cleanup_directories()
        if build_id in self._builds:
            build = self._builds[build_id]
        else:
            if build_id in self._results:
                del self._results[build_id]
            build = self._builds[build_id] = AgentBuild(
                self, build_id, context, processing, pack, end_url)
        self.start_waiting()
        return build

    # -------------------------------------------------------------------------
    def progress(self, build_id):
        """Return the progress of build.

        :param build_id: (string)
            Build ID.
        :return: (tuple)
            A tuple such as ``(<step>, <percent>, <message>)``.

        The step ``<step>`` is one of the following:

        * ``a_start``: starting
        * ``a_env``: importing processor environment
        * ``a_build``: building
        * ``a_warn``: a warning occurred
        * ``a_error``: an error occurred
        * ``a_fatal``: a fatal error occurred
        * ``a_stop``: stopping
        * ``a_end``: successfully completed
        * ``none``: unknown or not in progress build
        """
        self._cleanup()
        if build_id in self._builds and self._builds[build_id].result['log']:
            return self._builds[build_id].result['log'][-1][1:]
        elif build_id in self._results:
            return self._results[build_id]['log'][-1][1:]
        self.start_waiting()
        return 'none', 0, ''

    # -------------------------------------------------------------------------
    def stop(self, build_id):
        """Stop a build.

        :param build_id: (string)
            Build ID.
        """
        self._cleanup()
        if build_id in self._builds:
            self._builds[build_id].stop()
        return ''

    # -------------------------------------------------------------------------
    def result(self, build_id):
        """Return the result of build.

        :param build_id: (string)
            Build ID.
        :return: (dictionary)
            ``self._result`` or ``{'status': 'none'}``.

        The status ``<status>`` is one of the following:

        * ``a_stop``: stopped
        * ``a_fatal``: a fatal error occurred
        * ``a_end``: successfuly completed
        * ``none``: unknown build
        """
        self._cleanup()
        self._cleanup_directories()
        return self._results[build_id] if build_id in self._results \
               else {'status': 'none', 'log': [], 'message': ''}

    # -------------------------------------------------------------------------
    def start_waiting(self):
        """Start waiting builds."""
        # Check waiting builds
        running = 0
        waiting = []
        for build_id in self._builds:
            if self._builds[build_id].result['log'][0][1] == 'a_wait':
                waiting.append(build_id)
            elif not self._builds[build_id].stopped():
                running += 1
        if running >= self._concurrent or not waiting:
            return

        # Start waiting
        waiting = sorted(waiting,
            key=lambda build_id: self._builds[build_id].result['log'][0][0])
        for build_id in waiting[self._concurrent - running:]:
            self._builds[build_id].expire = time() + self.build_ttl
        for build_id in waiting[0:self._concurrent - running]:
            self._builds[build_id].start()

    # -------------------------------------------------------------------------
    def _cleanup(self):
        """Delete completed builds and expired results and kill long builds."""
        # Build -> result or stop
        now = time()
        for build_id in self._builds.keys():
            if self._builds[build_id].stopped():
                self._builds[build_id].result['expire'] = now + self.result_ttl
                self._results[build_id] = self._builds[build_id].result
                del self._builds[build_id]
            elif self._builds[build_id].expire < now:
                self._builds[build_id].stop()

        # Remove old results
        for build_id in self._results.keys():
            if now > self._results[build_id]['expire']:
                del self._results[build_id]

    # -------------------------------------------------------------------------
    def _cleanup_directories(self):
        """Remove old directories."""
        # Clean up buid path
        now = time()
        if exists(self.build_root):
            for name in listdir(self.build_root):
                path = join(self.build_root, name)
                if not name in self._builds and exists(path) \
                       and getatime(path) + self.result_ttl < now:
                    shutil.rmtree(path)

        # Clean up buidspace path
        if self.buildspace_root is not None and exists(self.buildspace_root):
            for name in listdir(self.buildspace_root):
                path = join(self.buildspace_root, name)
                if exists(path) \
                       and getatime(path) + self._buildspace_ttl < now:
                    shutil.rmtree(path)


# =============================================================================
class AgentBuild(object):
    """This class manages one local build.

    ``self.result`` is a dictionary with the following keys: ``status``,
    ``message``, ``log``. Log entry is a list of tuples such as ``(<timestamp>,
    <step>, <percent>, <message>)``. Engine can add keys like: ``files``,
    ``values``.
    """
    # pylint: disable = I0011, R0902, R0913

    # -------------------------------------------------------------------------
    def __init__(self, build_manager, build_id, context, processing, pack,
                 end_url):
        """Constructor method.

        :param build_manager: (:class:`AgentBuildManager` instance)
            Application :class:`AgentBuildManager` object.
        :param build_id: (string)
            Build ID.
        :param context: (dictionary)
            See :class:`~publiforge.lib.build.front.FrontBuildManager`
            :meth:`~publiforge.lib.build.front.FrontBuildManager.call`
            method.
        :param processing: (dictionary)
            A processing dictionary.
        :param pack: (dictionary)
            A pack dictionary.
        :param end_url: (string, optional)
            URL to call to complete the build.
        """
        self.build_id = build_id
        self.context = context
        self.path = join(build_manager.build_root, camel_case(build_id))
        self.data_path = context['local'] and build_manager.storage_root \
                         or build_manager.buildspace_root
        self.result = {'status': 'none',
            'log': [(time(), 'a_wait', 1, self._translate(_('waiting...')))],
            'message': ''}
        self.expire = time() + build_manager.build_ttl
        self.processing = processing
        self.pack = pack
        self._build_manager = build_manager
        self._end_url = end_url
        self._thread = None
        if not exists(self.path):
            makedirs(self.path)

    # -------------------------------------------------------------------------
    def start(self):
        """Start the processing."""
        # Already started
        if self._thread is not None:
            self.stopped(_('Already in progress'), 'a_warn')
            return
        self.result['log'] = [(
            time(), 'a_start', 1, self._translate(_('agent build startup')))]

        # Create build directory
        proc_path = join(self.path, 'Processor')
        if self._build_manager.develop and exists(proc_path):
            shutil.rmtree(proc_path)
        if not exists(proc_path):
            makedirs(proc_path)
            if not self._import_processor(self.processing['processor_id']):
                self.stopped(self.result['message'])
                return
            if not exists(join(self.path, 'Output')):
                makedirs(join(self.path, 'Output'))
            self.log(_('processor environment installed'), 'a_env', 100)

        # Build directly...
        if self._end_url is None:
            self._thread_engine()

        # ...or in a thread
        self._thread = Thread(target=self._thread_engine)
        self._thread.start()

    # -------------------------------------------------------------------------
    def stop(self):
        """Stop building."""
        if not self.stopped():
            self.result['status'] = 'a_stop'
            self.result['message'] = self._translate(_('Stop by user'))
            self.log(_('stopped'), 'a_stop', 100)

    # -------------------------------------------------------------------------
    def stopped(self, error=None, level='a_fatal'):
        """Check if there is a fatal error and if the build is stopped.

        :param error: (string, optional)
            Error message.
        :param level: (string, default='a_fatal')
            Error level: ``a_warn``, ``a_error`` or ``a_fatal``.
        :return: (boolean)
            ``True`` if it is stopped.
        """
        if error:
            if level == 'a_fatal':
                self.result['status'] = level
                self.result['message'] = self._translate(error)
            self.log(error, level, 100)

        return self.result['status'] in ('a_stop', 'a_fatal', 'a_end')

    # -------------------------------------------------------------------------
    def log(self, message, step=None, percent=None):
        """Append an entry to ``result['log']``.

        :param message: (string)
            Message to write in log.
        :param step: (string, optional)
            If not ``None``, progress is updated.
        :param percent: (int, optional)
            Percent of progress for step ``<step>``.
        """
        if percent is None:
            percent = self.result['log'][-1][2] if step is None else 0
        if step is None:
            step = self.result['log'][-1][1]

        self.result['log'].append(
            (time(), step, percent, self._translate(message)))

        if self.context['local'] and not 'request' in self.context:
            {'a_warn': LOG.warning, 'a_error': LOG.error,
             'a_fatal': LOG.critical
             }.get(step,  LOG.info)(self.result['log'][-1][3])

    # -------------------------------------------------------------------------
    def _thread_engine(self):
        """Action in a thread to launch engine."""
        # Find the engine
        tree = etree.parse(join(self.path, 'Processor', 'processor.xml'))
        module_name = tree.findtext('processor/engine').strip()
        try:
            module = __import__(module_name, fromlist=['Engine'])
        except ImportError, err:
            self.stopped('%s: %s' % (module_name, err))

        # Launch engine
        if not self.stopped():
            self.log(_('start engine'), 'a_build')
            self.expire = time() + self._build_manager.build_ttl
            module.Engine(self).start()
            if not self.stopped():
                self.log(_('agent build completed'), 'a_end', 100)
                self.result['status'] = 'a_end'

        # Announce the end to the front
        if self._end_url is None:
            self._build_manager.start_waiting()
            return
        if not self.context['local']:
            try:
                response = urllib.urlopen(self._end_url)
            except IOError, err:
                self._build_manager.start_waiting()
                return
            response.close()
        elif 'request' in self.context:
            request = self.context['request']
            request.registry['fbuild'].complete(
                request, self.build_id, self._end_url.split('/')[-1])
        self._build_manager.start_waiting()

    # -------------------------------------------------------------------------
    def _import_processor(self, processor_id):
        """Import processor, eventually with inheritance, in build directory.

        :param processor_id: (string)
            ID of processor to import.
        :return: (boolean)
            ``True`` if it succeeds.
        """
        # Find processor
        src_path = self._build_manager.processor_path(processor_id)
        if not src_path:
            self.stopped(_('Unknown processor "${p}"', {'p': processor_id}))
            return False

        # Read processor.xml file to check if other processor is needed
        ancestors = etree.parse(join(src_path, 'processor.xml'))\
                   .findall('processor/ancestors/ancestor')
        for ancestor in ancestors:
            if not self._import_processor(ancestor.text.strip()):
                return False

        # Copy processor
        copy_content(src_path, join(self.path, 'Processor'))
        return True

    # -------------------------------------------------------------------------
    def _translate(self, text):
        """Return ``text`` translated.

        :param text: (string)
            Text to translate.
        """
        return localizer(self.context['lang']).translate(text)



# $Id: __init__.py d9ce5e0005bd 2011/12/27 11:47:04 patrick $
# pylint: disable = I0011, R0904
"""Unit, integration, and functional testing."""

import unittest
import webtest
from os.path import join, dirname, normpath

from pyramid import testing


TEST_INI = normpath(join(dirname(__file__), '..', '..', 'development.ini'))


# =============================================================================
class ModelTestCase(unittest.TestCase):
    """Base ``TestCase`` class with SqlAlchemy session."""
    # pylint: disable = I0011, C0103

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set SqlAlchemy session."""
        from logging import WARNING
        from ..models import DBSession
        from ..scripts.pfpopulate import Populate, LOG as pfpopulate_log
        from ..lib.vcs.hg import LOG as hg_log
        populate = Populate(TEST_INI)
        pfpopulate_log.setLevel(WARNING)
        hg_log.setLevel(WARNING)
        populate.all(True)
        self.session = DBSession
        self.config = testing.setUp()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        self.session.remove()
        testing.tearDown()

    # -------------------------------------------------------------------------
    @classmethod
    def _make_request(cls):
        """Make a dummy request."""
        class DummyBreadcrumbs(object):
            """Dummy breadcrumb trail for :class:`DummyRequest`."""
            def add(self, title, length=0):
                """Add a crumb in breadcrumb trail."""
                pass

            @classmethod
            def current_path(cls):
                """Path of current page."""
                return '/'

        request = testing.DummyRequest()
        request.session['lang'] = 'en'
        request._LOCALE_ = request.session['lang']
        request.breadcrumbs = DummyBreadcrumbs()
        return request


# =============================================================================
class FunctionalTestCase(unittest.TestCase):
    """Base ``TestCase`` class with ``testapp`` attribute."""
    # pylint: disable = I0011, C0103

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.testapp = TestApp('config:%s' % TEST_INI, relative_to='.')

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models import DBSession
        del self.testapp
        DBSession.remove()


# =============================================================================
class TestApp(webtest.TestApp):
    """An extended ``TestApp`` class with login method."""

    # -------------------------------------------------------------------------
    def login(self, user_login):
        """Quick login.

        :param user_login: (string)
            User login.
        """
        csrf = self.get('/login').form.get('_csrf').value
        params = {'_csrf': csrf, 'login': user_login}
        self.post('/login_test', params, status=302)



# $Id: __init__.py a8f1b45b2ed8 2011/12/10 10:19:48 patrick $
"""The :func:`.main` function is called when the ``pserve`` command is invoked
against this application.
"""

from pyramid.config import Configurator
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy

from .initialize import Initialize, perm_group_finder, RootFactory


# =============================================================================
def main(global_config, **settings):
    """This is the main function. It returns a Pyramid WSGI application.

    :param global_config: (dictionary)
        Default section of INI file.
    :param settings: (dictionary)
        Application settings of [app:publiforge] section of INI file.
    :return: (object)
        WSGI application.

    It creates an instance of :class:`~.initialize.Initialize` class and call
    its :meth:`~.initialize.Initialize.all` method to initialize the
    application.
    """
    authentication_policy = AuthTktAuthenticationPolicy(
        secret=settings.get('auth.key', 'seekrit'),
        cookie_name=settings.get('auth.cookie', 'PF_AUTH'),
        callback=perm_group_finder)
    authorization_policy = ACLAuthorizationPolicy()
    config = Configurator(settings=settings,
                          root_factory=RootFactory,
                          authentication_policy=authentication_policy,
                          authorization_policy=authorization_policy)

    Initialize(global_config, config).all()

    return config.make_wsgi_app()



# $Id: publidoc2html_final.py 1292397d8520 2011/12/30 17:33:03 patrick $
"""LePrisme finalization script."""

from os.path import basename

from publiforge.lib.utils import _


# =============================================================================
def main(engine, output):
    """Finalization.

    :param engine: (Engine object)
        Engine object on which it depends.
    :param output: (string)
        Full path to output directory.
    """
    engine.build.log(
        _('dummy finalization of "${f}"', {'f': basename(output)}))



# $Id: processing.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
# pylint: disable = I0011, C0322
"""Processing view callables."""

from os.path import join, normpath
from colander import Mapping, SchemaNode
from colander import String, Length, OneOf
from webhelpers.html import literal

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPForbidden

from ..lib.utils import _, FILE_TYPES, has_permission
from ..lib.xml import export_configuration, local_text
from ..lib.form import Form
from ..lib.views import get_action, current_storage, current_project
from ..lib.views import current_processing, file_infos, variable_schema
from ..lib.views import describe_var
from ..lib.renderer import TabSet
from ..views.pack import FILE_TYPE_LABELS, file_infos
from ..models import LABEL_LEN, DESCRIPTION_LEN, PATH_LEN
from ..models import DBSession, close_dbsession
from ..models.processors import Processor
from ..models.projects import ProjectProcessing, ProjectProcessingFile
from ..models.projects import ProjectProcessingVariable


PROCESSING_SETTINGS_TABS = (
    _('Description'), _('Variables'), _('Files'), _('Output'))


# =============================================================================
class ProcessingView(object):
    """Class to manage processings."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='processing_view',
                 renderer='../Templates/prc_view.pt')
    def view(self):
        """Display processing settings."""
        # Permission
        project = current_project(self._request)
        i_editor = has_permission(self._request, 'prj_editor') \
            or project['perm'] == 'editor'

        # Environment
        processing, processor = current_processing(
            self._request, container=True)
        variables = dict([(k.name, k) for k in processing.variables])
        tab_set = TabSet(self._request, PROCESSING_SETTINGS_TABS)
        output = processing.output and current_storage(
            self._request, processing.output.partition('/')[0])[0] or None

        # Action
        action, description = self._action(
            None, processing, processor, variables)
        if action == 'exp!':
            return  export_configuration((processing.xml(processor),))

        # Breadcrumbs
        self._request.breadcrumbs.add(
            _('Processing'), replace=self._request.route_path(
            'processing_edit', project_id=project['project_id'],
            processing_id=processing.processing_id))

        return {
            'form': Form(self._request), 'tab_set': tab_set,
            'file_type_labels': FILE_TYPE_LABELS, 'FILE_TYPES': FILE_TYPES,
            'project': project, 'processing': processing,
            'processor': processor, 'variables': variables,
            'files': file_infos(self._request, processing), 'output': output,
            'description': description, 'i_editor': i_editor}

    # -------------------------------------------------------------------------
    @view_config(route_name='processing_create',
                 renderer='../Templates/prc_edit.pt')
    def create(self):
        """Create a processing."""
        # Permission
        project = current_project(self._request)
        if not has_permission(self._request, 'prj_editor') \
               and project['perm'] != 'editor':
            raise HTTPForbidden()

        # Environment
        processor_id = self._request.params.get('processor_id')
        processor_labels = Processor.labels(self._request)
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='label',
            validator=Length(min=2, max=LABEL_LEN)))
        schema.add(SchemaNode(String(), name='description', missing='',
            validator=Length(max=DESCRIPTION_LEN)))
        schema.add(SchemaNode(String(), name='processor_id',
            validator=OneOf(processor_labels.keys())))
        form = Form(self._request, schema=schema)
        tab_set = TabSet(self._request, PROCESSING_SETTINGS_TABS)

        # Action
        action = get_action(self._request)[0]
        description = None
        if action == 'des!' and processor_id:
            description = Processor.description(self._request, processor_id)
            description = literal('<b>%s (%s)</b><br/>%s' %
                (processor_labels[processor_id], processor_id, description))
        elif action == 'sav!' and form.validate():
            processing = self._create(project['project_id'], form.values)
            if processing is not None:
                if 'project' in self._request.session:
                    del self._request.session['project']
                self._request.breadcrumbs.pop()
                return HTTPFound(self._request.route_path('processing_edit',
                    project_id=processing.project_id,
                    processing_id=processing.processing_id))

        self._request.breadcrumbs.add(_('Processing creation'))
        return {
            'form': form, 'tab_set': tab_set, 'project': project,
            'processing': None, 'processor_labels': processor_labels,
            'description': description}

    # -------------------------------------------------------------------------
    @view_config(route_name='processing_edit',
                 renderer='../Templates/prc_edit.pt')
    def edit(self):
        """Edit a processing."""
        # Permission
        project = current_project(self._request)
        if not has_permission(self._request, 'prj_editor') \
               and project['perm'] != 'editor':
            raise HTTPForbidden()

        # Environment
        processing, processor = current_processing(
            self._request, container=True)
        variables = dict([(k.name, k) for k in processing.variables])
        form, tab_set = self._settings_form(processing, processor, variables)
        output = processing.output and current_storage(
            self._request, processing.output.partition('/')[0])[0] or None

        # Action
        action, description = self._action(
            form, processing, processor, variables, True)
        view_path = self._request.route_path('processing_view',
            project_id=processing.project_id,
            processing_id=processing.processing_id)
        if action == 'view':
            return HTTPFound(view_path)
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        self._request.breadcrumbs.add(_('Processing'), replace=view_path)

        return {
            'form': form, 'tab_set': tab_set, 'action': action,
            'file_type_labels': FILE_TYPE_LABELS, 'FILE_TYPES': FILE_TYPES,
            'project': project, 'processing': processing,
            'processor': processor, 'variables': variables,
            'files': file_infos(self._request, processing), 'output': output,
            'description': description}

    # -------------------------------------------------------------------------
    def _settings_form(self, processing, processor, variables):
        """Return a processing settings form.

        :param processing: (:class:`~..models.projects.ProjectProcessing`
            instance) Current processing object.
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param variables: (dictionary)
            Variables of current processing.
        :return: (tuple)
             A tuple such as ``(form, tab_set)``
        """
        schema, defaults = variable_schema(processor, variables)
        schema.add(SchemaNode(String(), name='label',
            validator=Length(min=2, max=LABEL_LEN)))
        schema.add(SchemaNode(String(), name='description',
            validator=Length(max=DESCRIPTION_LEN), missing=''))
        defaults['label'] = processing.label
        defaults['description'] = processing and processing.description

        if not processing.output \
               and processor.findtext('processor/output') is not None:
            processing.output = processor.findtext('processor/output').strip()
            DBSession.commit()
        if processing.output:
            schema.add(SchemaNode(String(), name='output',
                validator=Length(max=PATH_LEN), missing=''))
            defaults['output'] = processing.output.partition('/')[2]

        for item in processing.files:
            if item.file_type == 'template':
                schema.add(SchemaNode(String(), name='template_%s' % item.path,
                    validator=Length(max=PATH_LEN)))
                defaults['template_%s' % item.path] = item.target

        return (Form(self._request, schema=schema, defaults=defaults),
                TabSet(self._request, PROCESSING_SETTINGS_TABS))

    # -------------------------------------------------------------------------
    def _action(self, form, processing, processor, variables,
                show_processor_defaults=False):
        """Return current action and help message.

        :param form: (:class:`~..lib.form.Form` instance)
            Current form.
        :param processing:
            (:class:`~..models.projects.ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param variables: (dictionary)
            Variables of current processing.
        :param show_processor_defaults: (boolean, default=False)
            Show the processor default value if any.
        :return: (tuple)
            A tuple such as ``(action, description)``.
        """
        action = get_action(self._request)[0]
        description = None

        if action == 'des!':
            description = Processor.description(self._request, xml=processor)
            description = literal('<b>%s </b><br/>%s' % (local_text(
                processor, 'processor/label', self._request), description))
        elif action[0:4] == 'des!':
            description = describe_var(self._request, processor, variables,
                action[4:], show_processor_defaults)
        elif action[0:4] == 'rst!':
            name = action[4:]
            var = processor.xpath(
                'processor/variables/group/var[@name="%s"]' % name)
            if len(var) == 1 and var[0].findtext('default') is not None:
                value = var[0].findtext('default').strip()
                form.values[name] = value
                form.static(name)
                form.values['%s_def' % name] = value
                form.static('%s_def' % name)
        elif action[0:4] == 'rmv!':
            DBSession.query(ProjectProcessingFile).filter_by(
                project_id=processing.project_id,
                processing_id=processing.processing_id,
                file_type=action[4:].partition('_')[0],
                path=action[4:].partition('_')[2]).delete()
            DBSession.commit()
        elif action == 'sav!' and form.validate() \
                 and self._save(processing, processor, variables, form.values):
            action = 'view'

        return action, description

    # -------------------------------------------------------------------------
    def _create(self, project_id, values):
        """Create a record in ``projects_processings`` table.

        :param project_id: (string)
            Project ID.
        :param values: (dictionary)
            Values to record.
        :return:
            (:class:`~..models.projects.ProjectProcessing`  instance)
        """
        # Check unicity and create processing
        label = u' '.join(values['label'].split())
        processing = DBSession.query(ProjectProcessing).filter_by(
            project_id=project_id, label=label).first()
        if processing is not None:
            self._request.session.flash(
                _('This processing already exists.'), 'alert')
            return
        processing = ProjectProcessing(
            project_id, label, values['description'], values['processor_id'])

        # Create variables
        processor = self._request.registry['fbuild'].processor(
            self._request, values['processor_id'])
        if processor.findtext('processor/output') is not None:
            processing.output = processor.findtext('processor/output').strip()
        for var in processor.findall('processor/variables/group/var'):
            if var.findtext('default') is not None:
                value = var.findtext('default').strip()
                processing.variables.append(
                    ProjectProcessingVariable(var.get('name'), value, value))

        DBSession.add(processing)
        DBSession.commit()

        return processing

    # -------------------------------------------------------------------------
    def _save(self, processing, processor, variables, values):
        """Save a processing settings.

        :param processing:
            (:class:`~..models.projects.ProjectProcessing` instance) Processing
            to update.
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param variables: (dictionary)
            Variables of current processing.
        :param values: (dictionary)
            Form values.
        :return: (boolean)
            ``True`` if succeeds.
        """
        processing.label = ' '.join(values['label'].split())
        processing.description = ' '.join(values['description'].split())

        if processing.output:
            processing.output = normpath(join(
                processing.output.partition('/')[0],
                values['output'].replace('..', '')))

        for var in processor.findall('processor/variables/group/var'):
            name = var.get('name')
            if name in variables:
                variables[name].value = values[name]
                variables[name].default = values['%s_def' % name]
                variables[name].visible = values['%s_see' % name]
            else:
                processing.variables.append(ProjectProcessingVariable(
                    name, values[name], values['%s_def' % name],
                    values['%s_see' % name]))

        for item in processing.files:
            if item.file_type == 'template':
                item.target = values['template_%s' % item.path]

        DBSession.commit()
        if 'build' in self._request.session:
            del self._request.session['build']

        return True



# $Id: test_models_projects.py 91a251074d63 2012/03/18 23:07:53 patrick $
# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``models.projects`` classes."""

import unittest

from ..tests import ModelTestCase


# =============================================================================
class UnitTestModelsProjectsProject(ModelTestCase):
    """Unit test class for ``models.projects.Project``."""
    # pylint: disable = I0011, C0103

    _project_id = 1000

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.projects import Project
        self.session.query(Project)\
            .filter_by(project_id=self._project_id).delete()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make an ``Project`` object."""
        from ..models.projects import Project
        project = Project(u'Les Misérables', u'Un roman de Virtor HUGO.')
        project.project_id = self._project_id
        return project

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.projects.Project.__init__]"""
        from ..models.projects import Project
        project1 = self._make_one()
        self.session.add(project1)
        self.session.commit()
        project2 = self.session.query(Project)\
                     .filter_by(project_id=project1.project_id).first()
        self.assertNotEqual(project2, None)
        self.assertEqual(project2.project_id, project1.project_id)
        self.assertEqual(project2.label, project1.label)
        self.assertEqual(project2.status, project1.status)
        self.assertEqual(project2.deadline, project1.deadline)

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.projects.Project.__repr__]"""
        project = self._make_one()
        info = repr(project)
        self.assertTrue('<Project(' in info)
        self.assertTrue('%d' % self._project_id in info)
        self.assertTrue("'%s'" % project.label.encode('utf8') in info)
        self.assertTrue("'%s'" % project.status in info)


# =============================================================================
class UnitTestModelsProjectsProjectUser(unittest.TestCase):
    """Unit test class for ``models.projects.ProjectUser``."""

    _project_id = 1000
    _user_id = 1000

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``ProjectUser`` object."""
        from ..models.projects import ProjectUser
        return ProjectUser(self._project_id, self._user_id)

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.projects.ProjectUser.__repr__]"""
        project_user = self._make_one()
        info = repr(project_user)
        self.assertTrue('<ProjectUser(' in info)
        self.assertTrue('%d' % self._project_id in info)
        self.assertTrue('%d' % self._user_id in info)


# =============================================================================
class IntegrationTestModelsProjects(ModelTestCase):
    """Integration test class for ``models.projects``."""
    # pylint: disable = I0011, C0103

    _project_id = 1000
    _user_id = 1000

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.users import User
        from ..models.projects import Project
        self.session.query(User).filter_by(user_id=self._user_id).delete()
        self.session.query(Project)\
            .filter_by(project_id=self._project_id).delete()
        self.session.commit()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_project(self):
        """Make a ``Project`` object."""
        # pylint: disable = I0011, E1101
        from ..models.projects import Project
        project = Project(u'Les Misérables', u'Un roman de Victor HUGO.')
        project.project_id = self._project_id
        return project

    # -------------------------------------------------------------------------
    def _make_user(self):
        """Make an ``User`` object."""
        from ..models.users import User
        user = User({'auth.key': 'seekrit'}, '_test_user', status='active',
                    password='mypassword', name=u'Marc HINDÉLAIBILE',
                    email='test@prismallia.fr', lang='en')
        user.user_id = self._user_id
        return user

    # -------------------------------------------------------------------------
    def test_remove_user(self):
        """[i:models.projects] remove user"""
        from ..models.users import User
        from ..models.projects import ProjectUser
        project = self._make_project()
        user = self._make_user()
        self.session.add(user)
        project.users.append(ProjectUser(project.project_id, self._user_id))
        self.session.add(project)
        self.session.commit()
        self.assertEqual(len(project.users), 1)
        project_user = self.session.query(ProjectUser).filter_by(
            project_id=project.project_id, user_id=self._user_id).first()
        self.assertFalse(project_user.in_menu)
        self.session.query(User).filter_by(user_id=self._user_id).delete()
        self.session.commit()
        self.assertEqual(len(project.users), 0)



# $Id: security.py de5826a61f18 2012/02/27 17:59:57 patrick $
# pylint: disable = I0011, C0322
"""Security view callables."""

from datetime import datetime, date
from colander import Mapping, SchemaNode, String, Boolean, Length

from pyramid.view import view_config
from pyramid.view import notfound_view_config, forbidden_view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotAcceptable
from pyramid.security import forget, remember, authenticated_userid
from pyramid.security import NO_PERMISSION_REQUIRED

from ..lib.utils import _, hash_sha
from ..lib.form import button, Form
from ..models import DBSession
from ..models.users import User


# =============================================================================
def login(request):
    """This view renders a login form and processes the post checking
    credentials.
    """
    # Create form
    schema = SchemaNode(Mapping())
    schema.add(SchemaNode(String(), name='login', validator=Length(min=2)))
    schema.add(SchemaNode(String(), name='password', validator=Length(min=8)))
    schema.add(SchemaNode(Boolean(), name='remember', missing=False))
    came_from = request.params.get('came_from') \
        or (request.url != request.route_url('login') and request.url) \
        or request.route_url('home')
    form = Form(request, schema=schema, defaults={'came_from': came_from})

    # Validate form
    # pylint: disable = I0011, E1103
    if form.validate():
        user_login = form.values['login']
        password = form.values['password']
        user = connect(request, user_login, password)
        if not isinstance(user, int):
            user.setup_environment(request)
            max_age = (form.values['remember']
                and request.registry.settings.get('auth.remember', '5184000'))\
                or None
            DBSession.close()
            return HTTPFound(location=came_from,
                headers=remember(request, user.user_id, max_age=max_age))
        request.session.flash({
            1: _('ID or password is incorrect.'),
            2: _('Your account is not active.'),
            3: _('Your account has expired.'),
            4: _('Your IP is rejected.')}[user], 'alert')

    DBSession.close()
    return {'form': form}


# =============================================================================
@view_config(route_name='login_test', permission=NO_PERMISSION_REQUIRED)
def login_test(request):
    """Quick log in for test only."""
    user_login = request.params.get('login')
    if not 'paste.testing' in request.environ or user_login is None:
        DBSession.close()
        return HTTPFound(location=request.route_path('login'))
    # pylint: disable = I0011, E1103
    user = connect(request, user_login)
    user.setup_environment(request)
    DBSession.close()
    return HTTPFound('/robots.txt',  headers=remember(request, user.user_id))


# =============================================================================
@view_config(route_name='logout')
def logout(request):
    """This view will clear the credentials of the logged in user and redirect
    back to the login page.
    """
    request.session.clear()
    DBSession.close()
    return HTTPFound(
        location=request.route_path('login'), headers=forget(request))


# =============================================================================
@notfound_view_config(renderer='../Templates/error.pt')
@forbidden_view_config(renderer='../Templates/error.pt')
@view_config(context=HTTPNotAcceptable, renderer='../Templates/error.pt',
             permission=NO_PERMISSION_REQUIRED)
def error(request):
    """This view outputs an error message or redirects to login page."""
    status = request.exception.status_int
    if status in (401, 403) and authenticated_userid(request) is None:
        return HTTPFound(request.route_path(
            'login', _query=(('came_from', request.path),)))

    request.response.status = status
    request.breadcrumbs.add(_('Error ${status}', {'status': status}))
    DBSession.close()
    return {'button': button,
            'message': request.exception.comment or {
            403: _('Access was denied to this resource.'),
            404: _('The resource could not be found.'),
            406: _('The server could not comply with the request since it '
                   'is either malformed or otherwise incorrect.'),
            500: _('Internal Server Error.')
            }.get(status, request.exception.explanation)}


# =============================================================================
def connect(request, code, password=None):
    """If the user with ``code`` and ``password`` is authorized, it updates
    ``last_login`` field and returns an ``User`` object.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param code: (string)
        Login or ID of the user to connect.
    :param password: (string, optional)
        Clear password.
    :return: (:class:`~.models.users.User` instance or integer)
        The connected user or an error code.

    If ``password`` is ``None``, password checking is not performed.

    Values of the error code are:

    * 1: incorrect user code or password
    * 2: inactive account
    * 3: expired account
    * 4: forbidden IP
    """
    # Check user
    if isinstance(code, int) and password is None:
        user = DBSession.query(User).filter_by(user_id=code).first()
    else:
        code = code.strip().lower()
        user = DBSession.query(User).filter_by(login=code).first()
    if user is None:
        return 1
    if password is not None:
        key = request.registry.settings['auth.key']
        if hash_sha(password.strip(), key) != user.password:
            return user.password is None and 2 or 1
    if user.status != 'active':
        return 2 if user.status == 'inactive' else 1
    if user.expiration and user.expiration < date.today():
        return 3
    if user.restrict_ip and not (request.environ['REMOTE_ADDR']
            in [str(k.ip) for k in user.ips]):
        return 4

    # Update last login date in database
    user.last_login = datetime.now()
    DBSession.commit()
    return user



# $Id: test_lib_xml.py 68b0cb3c67d5 2011/12/18 18:50:45 patrick $
# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``lib.xml`` classes and functions."""

import unittest
from os.path import join, dirname

XML_FILE = join(
    dirname(__file__), '..', 'Processors', 'PublidocValid', 'processor.xml')
INI_FILE = join(
    dirname(__file__), '..', 'Processors', 'Publidoc2Html', 'leprisme.ini')


# =============================================================================
class UnitTestLibUtilsLoad(unittest.TestCase):
    """Unit test class for ``lib.xml.load``."""

    # -------------------------------------------------------------------------
    @classmethod
    def _call_fut(cls, filename, relaxngs=None, data=None):
        """Call function under test."""
        from ..lib.xml import load
        if relaxngs is None:
            relaxngs = {'publiforge':
                join(dirname(__file__), '..', 'RelaxNG', 'publiforge.rng')}
        return load(filename, relaxngs, data)

    # -------------------------------------------------------------------------
    def test_not_well_formed(self):
        """[u:lib.xml.load] not well-formed XML"""
        tree = self._call_fut(INI_FILE)
        self.assertTrue(isinstance(tree, basestring))
        self.assertTrue('not well-formed' in tree)

    # -------------------------------------------------------------------------
    def test_invalid(self):
        """[u:lib.xml.load] invalid XML"""
        with open(XML_FILE, 'r') as hdl:
            data = hdl.read().replace('engine', 'motor')
        tree = self._call_fut(XML_FILE, data=data)
        self.assertTrue(isinstance(tree, basestring))
        self.assertTrue('RELAXNG_ERR_ELEMWRONG' in tree)

    # -------------------------------------------------------------------------
    def test_valid_xml(self):
        """[u:lib.xml.load] valid XML"""
        tree = self._call_fut(XML_FILE)
        self.assertFalse(isinstance(tree, basestring))



# $Id: projects.py 86b088d85a5a 2012/03/21 22:20:47 patrick $
"""SQLAlchemy-powered model definition for projects."""

from lxml import etree
from textwrap import fill
from datetime import datetime
from sqlalchemy import Column, ForeignKey, types
from sqlalchemy.schema import ForeignKeyConstraint, UniqueConstraint
from sqlalchemy.orm import relationship

from ..lib.utils import _, normalize_name
from . import ID_LEN, LABEL_LEN, DESCRIPTION_LEN, PATH_LEN, VALUE_LEN
from . import Base, DBSession
from .users import User
from .groups import Group


PROJECT_STATUS = {'draft': _('Draft'), 'active': _('Active'),
                  'archived': _('Archived')}
PROJECT_PERMS = {'editor': _('Leader'), 'user': _('User')}


# =============================================================================
class Project(Base):
    """SQLAlchemy-powered project model."""
    # pylint: disable = I0011, R0903, W0142

    __tablename__ = 'projects'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    project_id = Column(types.Integer, primary_key=True)
    label = Column(types.String(LABEL_LEN), unique=True, nullable=False)
    description = Column(types.String(DESCRIPTION_LEN))
    status = Column(types.Enum(*PROJECT_STATUS.keys(), name='prj_status_enum'),
                    nullable=False, default='active')
    deadline = Column(types.Date)
    users = relationship('ProjectUser', backref='project')
    groups = relationship('ProjectGroup')
    processings = relationship('ProjectProcessing')
    packs = relationship('ProjectPack')

    # -------------------------------------------------------------------------
    def __init__(self, label, description=None, status=None, deadline=None):
        """Constructor method."""
        self.label = u' '.join(label.strip().split())[0:LABEL_LEN]
        if description:
            self.description = \
                u' '.join(description.strip().split())[0:DESCRIPTION_LEN]
        self.status = status
        if deadline is not None:
            self.deadline = (isinstance(deadline, basestring)
                and datetime.strptime(deadline, '%Y-%m-%d')) or deadline

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return u"<Project({project_id}, '{label}', '{status}')>".format(
            project_id=self.project_id or -1, label=self.label,
            status=self.status).encode('utf8')

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, project_elt, error_if_exists=True):
        """Load a project from a XML file.

        :param project_elt: (:class:`lxml.etree.Element` instance)
            Project XML element.
        :param error_if_exists: (boolean, default=True)
            It returns an error if project already exists.
        :return: (:class:`pyramid.i18n.TranslationString` or ``None``)
            Error message or ``None``.
        """
        # Check if already exists
        label = u' '.join(
            project_elt.findtext('label').strip().split())[0:LABEL_LEN]
        project = DBSession.query(cls).filter_by(label=label).first()
        if project is not None:
            if error_if_exists:
                return _('Project "${l}" already exists.', {'l': label})
            return

        # Create project
        project = cls(project_elt.findtext('label'),
                      project_elt.findtext('description'),
                      project_elt.get('status', 'active'),
                      project_elt.findtext('deadline'))
        DBSession.add(project)
        DBSession.commit()

        # Append processings
        elt = project_elt.find('processings')
        if elt is not None:
            for child in elt.iterchildren(tag=etree.Element):
                ProjectProcessing.load(project.project_id, child, False)

        # Append packs
        elt = project_elt.find('packs')
        if elt is not None:
            for child in elt.iterchildren(tag=etree.Element):
                ProjectPack.load(project.project_id, child)

        # Add users
        done = []
        for item in project_elt.findall('members/member'):
            login = normalize_name(item.text)[0:ID_LEN]
            if not login in done:
                user = DBSession.query(User).filter_by(login=login).first()
                if user is not None:
                    project.users.append(ProjectUser(
                        project.project_id, user.user_id, item.get('in-menu'),
                        item.get('permission', 'user')))
                done.append(login)

        # Add groups
        done = []
        for item in project_elt.findall('members/member-group'):
            group_id = normalize_name(item.text)[0:ID_LEN]
            if group_id not in done:
                done.append(group_id)
                group = DBSession.query(Group).filter_by(
                    group_id=group_id).first()
                if group is not None:
                    project.groups.append(ProjectGroup(project.project_id,
                        group.group_id, item.get('permission', 'user')))

        DBSession.commit()

    # -------------------------------------------------------------------------
    def xml(self, request):
        """Serialize a project to a XML representation.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :return: (:class:`lxml.etree.Element`)
        """
        # Header
        project_elt = etree.Element('project')
        project_elt.set('status', self.status)
        etree.SubElement(project_elt, 'label').text = self.label
        if self.description:
            etree.SubElement(project_elt, 'description').text = fill(
                self.description, initial_indent='\n' + ' ' * 8,
                subsequent_indent=' ' * 8)
        if self.deadline:
            etree.SubElement(project_elt, 'deadline')\
                .text = self.deadline.isoformat()

        # Processings
        if self.processings:
            group_elt = etree.SubElement(project_elt, 'processings')
            for processing in self.processings:
                group_elt.append(etree.Comment(u'{0:~^64}'.format(
                    u' Processing: %s ' % processing.label)))
                processor = request.registry['fbuild'].processor(
                    request, processing.processor)
                group_elt.append(processing.xml(processor))

        # Packs
        if self.packs:
            group_elt = etree.SubElement(project_elt, 'packs')
            for pack in self.packs:
                group_elt.append(etree.Comment(u'{0:~^64}'.format(
                    u' Pack: %s ' % pack.label)))
                group_elt.append(pack.xml())

        # Team
        if self.users or self.groups:
            project_elt.append(etree.Comment(u'{0:~^66}'.format(' Team ')))
            group_elt = etree.SubElement(project_elt, 'members')
            for user in self.users:
                elt = etree.SubElement(group_elt, 'member')
                elt.text = user.user.login
                if user.perm != 'user':
                    elt.set('permission', user.perm or 'none')
                if user.in_menu:
                    elt.set('in-menu', 'true')
            for group in self.groups:
                elt = etree.SubElement(group_elt, 'member-group')
                elt.text = group.group_id
                if group.perm != 'user':
                    elt.set('permission', group.perm)

        return project_elt

    # -------------------------------------------------------------------------
    @classmethod
    def load_set(cls, root_elt, record, file_class):
        """Load set of files or variables.

        :param root_elt: (:class:`lxml.etree.Element` instance)
            Element to browse
        :param record: (:class:`~.models.projects.ProjectProcessing`
            or :class:`~.models.projects.ProjectPack` instance).
        :param file_class:
            (:class:`~.models.projects.ProjectProcessingFile`
            or :class:`~.models.projects.ProjectPackFile)`
            File class.
         """
        for child in root_elt.iterdescendants(tag=etree.Element):
            if child.tag == 'var':
                value = child.findtext('value') is not None \
                        and child.findtext('value').strip() or ''
                record.variables.append(ProjectProcessingVariable(
                    child.get('name'), value,
                    child.findtext('default') is not None
                    and child.findtext('default').strip() or value,
                    child.get('visible')))
            elif child.tag in ('file', 'resource', 'template'):
                record.files.append(file_class(
                    child.tag, child.text.strip(), child.get('to'),
                    child.get('visible')))

    # -------------------------------------------------------------------------
    @classmethod
    def xml_variables(cls, root_elt, processing, processor, with_default=True):
        """Read variable definitions in processor tree and fill ``variables``
        XML structure.

        :param root_elt: (:class:`lxml.etree.Element` instance)
            Element that linked the result.
        :param processing:
            (:class:`~ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param with_default: (boolean)
            Add default value in ``<var>``.
        """
        if not len(processing.variables):
            return

        # Browse variables
        group_elt = etree.SubElement(root_elt, 'variables')
        values = dict([(k.name, (k.value, k.default, k.visible))
                       for k in processing.variables])
        for group in processor.findall('processor/variables/group'):
            elt = group.xpath('label[@xml:lang="en"]')
            if len(elt):
                group_elt.append(
                    etree.Comment(u'{0:.^30}'.format(elt[0].text)))
            for var in group.findall('var'):
                value = values.get(var.get('name'), (None, None, False))
                elt = etree.SubElement(group_elt, 'var', name=var.get('name'))
                if value[2]:
                    elt.set('visible', 'true')
                etree.SubElement(elt, 'value').text = \
                    value[0] if value[0] is not None \
                    else (var.find('default') is not None
                          and var.find('default').text.strip()) \
                          or (var.get('type') == 'boolean' and 'false') or ''
                if with_default and value[1] is not None:
                    etree.SubElement(elt, 'default').text = value[1]

    # -------------------------------------------------------------------------
    @classmethod
    def xml_file_set(cls, root_elt, record, file_type, with_comment=True):
        """Save set of files in a XML object.

        :param root_elt: (:class:`lxml.etree.Element` instance)
            Element that linked the result.
        :param record: (:class:`~.models.projects.ProjectProcessing`
            or :class:`~.models.projects.ProjectPack` instance).
        :param file_type: ('file', 'resource' or 'template')
            File type.
        :param with_comment: (boolean, default=True)
            If ``True``, add a comment line before element.
         """
        items = [k for k in record.files if k.file_type == file_type]
        if len(items) == 0:
            return
        if with_comment:
            root_elt.append(
                etree.Comment('{0:.^32}'.format('%ss' % file_type)))
        group_elt = etree.SubElement(root_elt, '%ss' % file_type)
        for item in items:
            elt = etree.SubElement(group_elt, file_type)
            if hasattr(item, 'target') and item.target:
                elt.set('to', item.target)
            if hasattr(item, 'visible') \
                   and not (file_type == 'file' and item.visible) \
                   and not (file_type != 'file' and not item.visible):
                elt.set('visible', item.visible and 'true' or 'false')
            elt.text = item.path


# =============================================================================
class ProjectUser(Base):
    """SQLAlchemy-powered association table between ``Project`` and
    ``User``."""
    # pylint: disable = I0011, R0903, W0142

    __tablename__ = 'projects_users'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    project_id = Column(types.Integer, ForeignKey('projects.project_id',
        ondelete='CASCADE', onupdate='CASCADE'), primary_key=True)
    user_id = Column(types.Integer, ForeignKey('users.user_id',
        ondelete='CASCADE', onupdate='CASCADE'), primary_key=True)
    in_menu = Column(types.Boolean(), default=False)
    perm = Column(types.Enum(*PROJECT_PERMS.keys(), name='prj_perms_enum'))
    user = relationship('User')

    # -------------------------------------------------------------------------
    def __init__(self, project_id, user_id, in_menu=False, perm=None):
        """Constructor method."""
        self.project_id = project_id
        self.user_id = user_id
        self.in_menu = bool(in_menu)
        if perm != 'none':
            self.perm = perm

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectUser({project_id}, {user_id}, {in_menu}, '{perm}')>" \
               .format(project_id=self.project_id or -1, user_id=self.user_id,
                       in_menu=self.in_menu, perm=self.perm)


# =============================================================================
class ProjectGroup(Base):
    """SQLAlchemy-powered association table between ``Project`` and
    ``Group``."""
    # pylint: disable = I0011, W0142, R0903

    __tablename__ = 'projects_groups'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    project_id = Column(types.Integer, ForeignKey('projects.project_id',
        ondelete='CASCADE', onupdate='CASCADE'), primary_key=True)
    group_id = Column(types.String(ID_LEN), ForeignKey('groups.group_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    perm = Column(types.Enum(*PROJECT_PERMS.keys(), name='prj_perms_enum'),
                  default='user')

    # -------------------------------------------------------------------------
    def __init__(self, project_id, group_id, perm=None):
        """Constructor method."""
        self.project_id = project_id
        self.group_id = group_id.strip()[0:ID_LEN]
        self.perm = perm

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectGroup({project_id}, '{group_id}', '{perm}')>".format(
            project_id=self.project_id or -1, group_id=self.group_id,
            perm=self.perm)


# =============================================================================
class ProjectProcessing(Base):
    """SQLAlchemy-powered project processing model."""
    # pylint: disable = I0011, R0903

    __tablename__ = 'projects_processings'
    __table_args__ = (UniqueConstraint('project_id', 'label'),
                      {'mysql_engine': 'InnoDB'})

    project_id = Column(types.Integer, ForeignKey('projects.project_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    processing_id = Column(types.Integer, primary_key=True)
    label = Column(types.String(LABEL_LEN), nullable=False)
    description = Column(types.String(DESCRIPTION_LEN))
    processor = Column(types.String(ID_LEN), nullable=False)
    output = Column(types.String(PATH_LEN))
    variables = relationship('ProjectProcessingVariable')
    files = relationship('ProjectProcessingFile')

    # -------------------------------------------------------------------------
    def __init__(self, project_id, label, description, processor, output=None):
        """Constructor method."""
        self.project_id = project_id
        self.label = u' '.join(label.split()).strip()[0:LABEL_LEN]
        if description:
            self.description = \
                u' '.join(description.strip().split())[0:DESCRIPTION_LEN]
        self.processor = processor.strip()[0:ID_LEN]
        self.output = output and output.strip()[0:PATH_LEN] or ''

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectProcessing({project_id}, {processing_id}, '{label}',"\
               " '{processor}', '{output}')>".format(
            project_id=self.project_id or -1,
            processing_id=self.processing_id or -1, label=self.label,
            processor=self.processor, output=self.output)

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, project_id, processing_elt, check_if_exists=True):
        """Load a processing from a XML file.

        :param processing_elt: (:class:`lxml.etree.Element` instance)
            Processing XML element.
        :param check_if_exists: (boolean, default=True)
            Check if processing already exists before inserting.
        :return: (:class:`pyramid.i18n.TranslationString` or ``None``)
            Error message or ``None``.
        """
        if processing_elt is None:
            return _('nothing to do!')

        label = ' '.join(processing_elt.findtext('label').strip().split())
        if check_if_exists:
            processing = DBSession.query(cls).filter_by(
                project_id=project_id, label=label).first()
            if processing is not None:
                return _('Processing "${l}" already exists.', {'l': label})

        processing = cls(
            project_id, label,
            processing_elt.findtext('description'),
            processing_elt.findtext('processor').strip(),
            processing_elt.findtext('output') is not None
                and processing_elt.findtext('output').strip() or None)
        Project.load_set(processing_elt, processing, ProjectProcessingFile)

        DBSession.add(processing)
        DBSession.commit()

    # -------------------------------------------------------------------------
    def xml(self, processor):
        """Serialize a processing to a XML representation.

        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :return: (:class:`lxml.etree.Element`)
        """
        proc_elt = etree.Element('processing')
        etree.SubElement(proc_elt, 'label').text = self.label
        if self.description:
            etree.SubElement(proc_elt, 'description').text = fill(
                self.description, initial_indent='\n' + ' ' * 12,
                subsequent_indent=' ' * 12)
        etree.SubElement(proc_elt, 'processor').text = self.processor
        if processor is not None:
            Project.xml_variables(proc_elt, self, processor)
        Project.xml_file_set(proc_elt, self, 'resource')
        Project.xml_file_set(proc_elt, self, 'template')
        if self.output:
            proc_elt.append(etree.Comment('{0:.^32}'.format('Output')))
            etree.SubElement(proc_elt, 'output').text = self.output

        return proc_elt


# =============================================================================
class ProjectProcessingVariable(Base):
    """SQLAlchemy-powered project processing variable model."""
    # pylint: disable = I0011, R0903, W0142

    __tablename__ = 'projects_processings_variables'
    __table_args__ = (
        ForeignKeyConstraint(['project_id', 'processing_id'],
            ['projects_processings.project_id',
             'projects_processings.processing_id'],
            onupdate='CASCADE', ondelete='CASCADE'),
        {'mysql_engine': 'InnoDB'})
    project_id = Column(types.Integer, primary_key=True)
    processing_id = Column(types.Integer, primary_key=True)
    name = Column(types.String(ID_LEN), primary_key=True)
    value = Column(types.String(VALUE_LEN))
    default = Column(types.String(VALUE_LEN))
    visible = Column(types.Boolean, default=False)

    # -------------------------------------------------------------------------
    def __init__(self, name, value, default=None, visible=False):
        """Constructor method."""
        self.name = name.strip()[0:ID_LEN]
        self.value = \
            isinstance(value, basestring) and value[0:VALUE_LEN] or value
        self.default = \
            isinstance(default, basestring) and default[0:VALUE_LEN] or default
        self.visible = bool(visible)

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectProcessingVariable({project_id}, {processing_id}, "\
               "'{name}', '{value}', '{default}', {visible})>".format(
            project_id=self.project_id or -1,
            processing_id=self.processing_id or -1, name=self.name,
            value=self.value, default=self.default, visible=self.visible)


# =============================================================================
class ProjectProcessingFile(Base):
    """SQLAlchemy-powered project processing file model."""
    # pylint: disable = I0011, R0903

    __tablename__ = 'projects_processings_files'
    __table_args__ = (
        ForeignKeyConstraint(['project_id', 'processing_id'],
            ['projects_processings.project_id',
             'projects_processings.processing_id'],
            onupdate='CASCADE', ondelete='CASCADE'),
        {'mysql_engine': 'InnoDB'})
    project_id = Column(types.Integer, primary_key=True)
    processing_id = Column(types.Integer, primary_key=True)
    file_type = Column(types.Enum('resource', 'template',
        name='prjprcfil_type_enum'), primary_key=True)
    path = Column(types.String(PATH_LEN), primary_key=True)
    target = Column(types.String(PATH_LEN))

    # -------------------------------------------------------------------------
    def __init__(self, file_type, path, target=None, visible=None):
        """Constructor method."""
        # pylint: disable = I0011, W0613
        self.file_type = file_type
        self.path = path.strip()[0:PATH_LEN]
        self.target = (target and target.strip()[0:PATH_LEN]) \
            or (file_type == 'template' and
                self.path.partition('/')[2][0:PATH_LEN]) or None

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectProcessingFile({project_id}, {processing_id}, "\
               "'{file_type}', '{path}', '{target}')>".format(
            project_id=self.project_id or -1,
            processing_id=self.processing_id or -1, file_type=self.file_type,
            path=self.path, target=self.target)


# =============================================================================
class ProjectPack(Base):
    """SQLAlchemy-powered project pack model."""
    # pylint: disable = I0011, R0903

    __tablename__ = 'projects_packs'
    __table_args__ = (UniqueConstraint('project_id', 'label'),
                      {'mysql_engine': 'InnoDB'})

    project_id = Column(types.Integer, ForeignKey('projects.project_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    pack_id = Column(types.Integer, primary_key=True)
    label = Column(types.String(LABEL_LEN), nullable=False)
    description = Column(types.String(DESCRIPTION_LEN))
    recursive = Column(types.Boolean, default=False)
    files = relationship('ProjectPackFile')

    # -------------------------------------------------------------------------
    def __init__(self, project_id, label, description=None, recursive=False):
        """Constructor method."""
        self.project_id = project_id
        self.label = u' '.join(label.strip().split())[0:LABEL_LEN]
        if description:
            self.description = \
                u' '.join(description.strip().split())[0:DESCRIPTION_LEN]
        self.recursive = bool(recursive)

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectPack({project_id}, {pack_id}, '{label}', "\
               "{recursive})>".format(
            project_id=self.project_id or -1, pack_id=self.pack_id or -1,
            label=self.label, recursive=self.recursive).encode('utf8')

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, project_id, pack_elt):
        """Load a pack from a XML file.

        :param pack_elt: (:class:`lxml.etree.Element` instance)
            Pack XML element.
        """
        pack = cls(project_id, pack_elt.findtext('label'),
                   pack_elt.findtext('description'),
                   pack_elt.get('recursive'))
        Project.load_set(pack_elt, pack, ProjectPackFile)

        DBSession.add(pack)
        DBSession.commit()

    # -------------------------------------------------------------------------
    def xml(self):
        """Serialize a pack to a XML representation.

        :return: (:class:`lxml.etree.Element`)
        """
        pack_elt = etree.Element('pack')
        if self.recursive:
            pack_elt.set('recursive', 'true')
        etree.SubElement(pack_elt, 'label').text = self.label
        if self.description:
            etree.SubElement(pack_elt, 'description').text = fill(
                self.description, initial_indent='\n' + ' ' * 12,
                subsequent_indent=' ' * 12)
        Project.xml_file_set(pack_elt, self, 'file', False)
        Project.xml_file_set(pack_elt, self, 'resource', False,)
        Project.xml_file_set(pack_elt, self, 'template', False)

        return pack_elt


# =============================================================================
class ProjectPackFile(Base):
    """SQLAlchemy-powered project pack file model."""
    # pylint: disable = I0011, R0903

    __tablename__ = 'projects_packs_files'
    __table_args__ = (
        ForeignKeyConstraint(['project_id', 'pack_id'],
            ['projects_packs.project_id', 'projects_packs.pack_id'],
            onupdate='CASCADE', ondelete='CASCADE'),
        {'mysql_engine': 'InnoDB'})
    project_id = Column(types.Integer, primary_key=True)
    pack_id = Column(types.Integer, primary_key=True)
    file_type = Column(types.Enum('file', 'resource', 'template',
        name='prjpckfil_type_enum'), primary_key=True)
    path = Column(types.String(PATH_LEN), primary_key=True)
    target = Column(types.String(PATH_LEN))
    visible = Column(types.Boolean())

    # -------------------------------------------------------------------------
    def __init__(self, file_type, path, target=None, visible=None):
        """Constructor method."""
        self.file_type = file_type
        self.path = path.strip()[0:PATH_LEN]
        self.target = (target and target.strip()[0:PATH_LEN]) \
            or (file_type == 'template'
                and self.path.partition('/')[2][0:PATH_LEN]) or None
        self.visible = (visible is None and file_type == 'file' and True) \
                       or bool(visible)

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<ProjectPackFile({project_id}, {pack_id}, '{file_type}',"\
               "'{path}', {visible}, '{target}')>".format(
            project_id=self.project_id or -1, pack_id=self.pack_id or -1,
            file_type=self.file_type, path=self.path, visible=self.visible,
            target=self.target)
